# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Fleet module
"""

from . import util
from .air_model import AirModel


class Fleet(AirModel):
    """
    Manage a Fleet

    ### delete
    Delete the fleet. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the fleet

    ### refresh
    Syncs the fleet with all values returned by the API

    ### update
    Update the fleet with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Fleet {self.name} {self.id}>'


class FleetApi:
    """High-level interface for the Fleet API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/fleet/'
        self.url = self.url.replace('v1', 'v2')

    def get(self, fleet_id, **kwargs):
        """
        Get an existing fleet

        Arguments:
            fleet_id (str): Fleet ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Fleet`](/docs/fleet)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.fleets.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Fleet fleet01 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{fleet_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Fleet(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing fleets

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.fleets.list()
        [<Fleet fleet01 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Fleet fleet02 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=dict)
        res = res.json().get('results')
        return [Fleet(self, **fleet) for fleet in res]

    @util.required_kwargs(['name', 'organization'])
    def create(self, **kwargs):
        # pylint: disable=line-too-long
        """
        Create a new fleet
        Arguments:
            name (str): Name of the fleet
            organization (`Organization` | str): `Organization` or ID associated with the fleet
            prefix (str, optional): Prefix for the fleet's IPv6 addresses
            prefix_length (int, optional): Length of the IPv6 address prefix
            gateway_ipv4 (str, optional): IPv4 gateway address for the fleet
            port_range (str, optional): Range of ports available on the fleet
            container_ipv4_network (str, optional): IPv4 network address for containers in the fleet
            container_prefix (int, optional): Prefix length for container IP addresses
            labels (list, optional): List of labels to assign to the fleet
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
            [`Fleet`](/docs/fleet)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.fleets.create(name='MyFleet', prefix='fd10:1:1::', prefix_length=48, gateway_ipv4='10.0.0.1', port_range='20000-40000', container_ipv4_network='172.16.0.0', container_prefix=16, organization=MyOrganization, labels=['label1', 'label2'])
        <Fleet MyFleet 3dadd54d-583c-432e-9383-a2b0b1d7f221>
        ```
        """  # pylint: enable=line-too-long
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Fleet(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Dict, Any

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints import Account
from air_sdk.v2.endpoints.organizations import Organization
from air_sdk.v2.air_model import AirModel, BaseEndpointApi


@dataclass(eq=False)
class Image(AirModel):
    id: str = field(repr=False)
    agent_enabled: bool = field(repr=False)
    archived: bool = field(repr=False)
    bios: Optional['Image'] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    bus: str = field(repr=False)
    console_support: Dict[str, Any] = field(repr=False)
    cpu_arch: str = field(repr=False)
    created: datetime = field(repr=False)
    default_username: Optional[str] = field(repr=False)
    default_password: Optional[str] = field(repr=False)
    features: Dict[str, Any] = field(repr=False)
    modified: Optional[datetime] = field(repr=False)
    mountpoint: Optional[str] = field(repr=False)
    minimum_resources: Dict[str, int] = field(repr=False)
    name: Optional[str]
    notes: Optional[str] = field(repr=False)
    organization: Optional[Organization] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    published: Optional[bool] = field(repr=False)
    provider: Optional[str] = field(repr=False)
    release_notes: Optional[str] = field(repr=False)
    simx: Optional[str] = field(repr=False)
    size: int = field(repr=False)
    upload_status: str = field(repr=False)
    uploader: Optional[Account] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    user_manual: Optional[str] = field(repr=False)
    version: str
    can_edit: bool = field(repr=False)
    organization_name: Optional[str]
    uploader_username: Optional[str] = field(repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return ImageEndpointApi


class ImageEndpointApi(mixins.ListApiMixin[Image], mixins.GetApiMixin[Image], BaseEndpointApi[Image]):
    API_PATH = 'images'
    model = Image



# SPDX-FileCopyrightText: Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Logger module
"""

import logging
import re


def _redact(record):
    """Redact any strings in the log message that match a sensitive pattern"""
    sensitive_patterns = [r'(password[\'\"]:\s?[\'\"]).*([\'\"])']
    for pattern in sensitive_patterns:
        record.msg = re.sub(pattern, r'\g<1>***\g<2>', record.msg)
    return record


air_sdk_logger = logging.getLogger('air_sdk')  # pylint: disable=invalid-name
air_sdk_logger.addHandler(logging.NullHandler())
air_sdk_logger.addFilter(_redact)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Optional

from air_sdk import const
from air_sdk.v2 import utils
from air_sdk.v2.client import Client

__all__ = ['const', 'AirApi', 'Client', 'utils', 'AirModelAttributeError']


from air_sdk.v2.exceptions import AirError, AirModelAttributeError

if TYPE_CHECKING:
    from air_sdk.v2.endpoints import (
        AccountEndpointApi,
        AnnouncementEndpointApi,
        ApiTokenEndpointApi,
        CloudInitEndpointApi,
        ImageEndpointApi,
        InterfaceEndpointApi,
        JobEndpointApi,
        LinkEndpointApi,
        MarketplaceDemoEndpointApi,
        MarketplaceDemoTagsEndpointApi,
        NodeEndpointApi,
        OrganizationEndpointApi,
        ResourceBudgetEndpointApi,
        ServiceEndpointApi,
        SimulationEndpointApi,
        SystemEndpointApi,
        UserConfigEndpointApi,
        WorkerEndpointApi,
    )


class AirApi:
    def __init__(
        self,
        api_url: str = const.DEFAULT_API_URL,
        username: Optional[str] = None,
        password: Optional[str] = None,
        bearer_token: Optional[str] = None,
        authenticate: bool = True,
    ):
        # Setup client session
        self.client = Client(api_url=api_url)
        if authenticate:
            self.client.authenticate(
                username=username,
                password=password,
                bearer_token=bearer_token,
            )

    @property
    def accounts(self) -> AccountEndpointApi:
        from .endpoints import AccountEndpointApi

        return AccountEndpointApi(self)

    @property
    def announcements(self) -> AnnouncementEndpointApi:
        from .endpoints import AnnouncementEndpointApi

        return AnnouncementEndpointApi(self)

    @property
    def api_tokens(self) -> ApiTokenEndpointApi:
        from .endpoints import ApiTokenEndpointApi

        return ApiTokenEndpointApi(self)

    @property
    def cloud_inits(self) -> CloudInitEndpointApi:
        from air_sdk.v2.endpoints import CloudInitEndpointApi

        return CloudInitEndpointApi(self)

    @property
    def images(self) -> ImageEndpointApi:
        from .endpoints import ImageEndpointApi

        return ImageEndpointApi(self)

    @property
    def interfaces(self) -> InterfaceEndpointApi:
        from .endpoints import InterfaceEndpointApi

        return InterfaceEndpointApi(self)

    @property
    def jobs(self) -> JobEndpointApi:
        from .endpoints import JobEndpointApi

        return JobEndpointApi(self)

    @property
    def links(self) -> LinkEndpointApi:
        from air_sdk.v2.endpoints import LinkEndpointApi

        return LinkEndpointApi(self)

    @property
    def marketplace_demos(self) -> MarketplaceDemoEndpointApi:
        from .endpoints import MarketplaceDemoEndpointApi

        return MarketplaceDemoEndpointApi(self)

    @property
    def marketplace_demo_tags(self) -> MarketplaceDemoTagsEndpointApi:
        from .endpoints import MarketplaceDemoTagsEndpointApi

        return MarketplaceDemoTagsEndpointApi(self)

    @property
    def nodes(self) -> NodeEndpointApi:
        from .endpoints import NodeEndpointApi

        return NodeEndpointApi(self)

    @property
    def organizations(self) -> OrganizationEndpointApi:
        from .endpoints import OrganizationEndpointApi

        return OrganizationEndpointApi(self)

    @property
    def resource_budgets(self) -> ResourceBudgetEndpointApi:
        from .endpoints import ResourceBudgetEndpointApi

        return ResourceBudgetEndpointApi(self)

    @property
    def services(self) -> ServiceEndpointApi:
        from .endpoints import ServiceEndpointApi

        return ServiceEndpointApi(self)

    @property
    def simulations(self) -> SimulationEndpointApi:
        from .endpoints import SimulationEndpointApi

        return SimulationEndpointApi(self)

    @property
    def systems(self) -> SystemEndpointApi:
        from .endpoints import SystemEndpointApi

        return SystemEndpointApi(self)

    @property
    def user_configs(self) -> UserConfigEndpointApi:
        from .endpoints import UserConfigEndpointApi

        return UserConfigEndpointApi(self)

    @property
    def workers(self) -> WorkerEndpointApi:
        from .endpoints import WorkerEndpointApi

        return WorkerEndpointApi(self)

    def set_connect_timeout(self, t_delta: timedelta) -> None:
        self.client.connect_timeout = int(t_delta.total_seconds())

    def set_read_timeout(self, t_delta: timedelta) -> None:
        self.client.read_timeout = int(t_delta.total_seconds())

    def set_page_size(self, n: int) -> None:
        """Set the page size of paginated responses."""
        if isinstance(n, int) and n > 0:
            self.client.pagination_page_size = n
        else:
            raise AirError('Pagination page size must be a positive integer.')



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Interface module
"""

from . import util
from .air_model import AirModel


class Interface(AirModel):
    """
    View an Interface

    ### json
    Returns a JSON string representation of the interface

    ### refresh
    Syncs the interface with all values returned by the API
    """

    _deletable = False
    _updatable = False

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Interface {self.name} {self.id}>'


class InterfaceApi:
    """High-level interface for the Interface API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/interface/'

    def get(self, interface_id, **kwargs):
        """
        Get an existing interface

        Arguments:
            interface_id (str): Interface ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Interface`](/docs/interface)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.interfaces.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Interface eth0 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{interface_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Interface(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing interfaces

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.interfaces.list()
        [<Interface eth0 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Interface eth1 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """
        # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Interface(self, **interface) for interface in res.json()]



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Account module
"""

from . import user_preference, util
from .air_model import AirModel


class Account(AirModel):
    """
    Manage an Account
    ### json
    Returns a JSON string representation of the account

    ### refresh
    Syncs the account with all values returned by the API
    """

    _deletable = False
    _updatable = False

    def __repr__(self):
        if self._deleted or not self.username:
            return super().__repr__()
        return f'<Account {self.username} {self.id}>'


class AccountApi:
    """High-level interface for the Account API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/account/'

    def get(self, account_id, **kwargs):
        """
        Get an existing account

        Arguments:
            account_id (str): Account ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Account`](/docs/account)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.accounts.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Account mrobertson@nvidia.com 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{account_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Account(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing accounts

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.accounts.list()
        [<Account mrobertson@nvidia.com c51b49b6-94a7-4c93-950c-e7fa4883591>, <Account nmitchell@nvidia.com 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """
        # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Account(self, **account) for account in res.json()]

    def preferences(self, **kwargs):
        """
        Returns your global account preferences

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`UserPreference`](/docs/userpreference)

        Example:
        ```
        >>> air.accounts.preferences()
        {"show": true}
        ```
        """
        res = self.client.get(f'{self.url}preferences/', params=kwargs)
        util.raise_if_invalid_response(res)
        return user_preference.UserPreference(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.air_model import AirModel, BaseEndpointApi
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class Announcement(AirModel):
    id: str = field(repr=False)
    severity: str
    created: datetime
    modified: datetime = field(repr=False)
    message: Optional[str] = field(repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return AnnouncementEndpointApi


class AnnouncementEndpointApi(
    mixins.ListApiMixin[Announcement],
    mixins.CreateApiMixin[Announcement],
    mixins.DeleteApiMixin,
    BaseEndpointApi[Announcement],
):
    API_PATH = 'announcements'
    model = Announcement

    @validate_payload_types
    def create(self, severity: str, message: str) -> Announcement:
        return super().create(severity=severity, message=message)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, fields, asdict, Field, is_dataclass, InitVar
from datetime import datetime
from functools import cached_property
from typing import (
    Union,
    Any,
    Dict,
    Mapping,
    TypeVar,
    ClassVar,
    Generic,
    Iterator,
    Type,
    get_type_hints,
    List,
    get_origin,
    cast,
    Optional,
    Tuple,
)
from uuid import UUID

from air_sdk.v2.exceptions import AirModelAttributeError, AirError
from air_sdk.v2.typing import is_optional_union, get_optional_arg, get_list_arg, is_union, type_check
from air_sdk.v2.utils import join_urls, iso_string_to_datetime, to_uuid, is_dunder, as_field

from air_sdk.v2 import AirApi

T = TypeVar('T')
TAirModel = TypeVar('TAirModel', bound='AirModel')
TAirModel_co = TypeVar('TAirModel_co', bound='AirModel', covariant=True)
TSupportedPrimitive = TypeVar('TSupportedPrimitive', int, str, float, bool)

PrimaryKey = Union[str, UUID]
SpecialField = Mapping[str, object]
DataDict = Dict[str, Any]


def _generate_special_field() -> SpecialField:
    """Returns a unique mapping reserved for assignment of `metadata` for special `BaseModel` fields."""
    return {'property': object()}


@dataclass(eq=False)
class BaseModel:
    __serializing__ = False  # A flag indicating if the instance is serializing

    def __eq__(self, other: Any) -> bool:
        if type(other) is type(self) and (pk := getattr(self, '__pk__', None)):
            return bool(pk == getattr(other, '__pk__', None))
        return bool(self is other)

    def dict(self) -> DataDict:
        try:
            self.__serializing__ = True
            result = asdict(self)
        finally:
            self.__serializing__ = False
        return result

    def json(self) -> str:
        from air_sdk.v2.endpoints.mixins import serialize_payload

        return serialize_payload(self.dict())


@dataclass(eq=False)
class AirModel(BaseModel, ABC):
    _api: InitVar  # type: ignore[type-arg]
    FIELD_FOREIGN_KEY: ClassVar[SpecialField] = _generate_special_field()
    FIELD_LAZY: ClassVar[SpecialField] = _generate_special_field()

    def __post_init__(self, _api: AirApi) -> None:
        self.__api__ = _api

    @property
    def primary_key_field(self) -> str:
        """Returns the name of the field representing the primary key."""
        return 'id'

    @property
    def __pk__(self) -> PrimaryKey:
        """
        Returns current model's primary key for API-related actions.
        """
        pk: PrimaryKey = getattr(self, self.primary_key_field)
        return pk

    def __getattribute__(self, name: str) -> Any:
        value = super().__getattribute__(name)

        # filter out dunder attributes (avoids recursion on upcoming `as_field` call)
        if value and not is_dunder(name):
            field = as_field(self, name)

            if field is not None:
                if self.__serializing__ and field.metadata == AirModel.FIELD_FOREIGN_KEY:
                    # When serializing ForeignKey fields use the pk instead of recursively serializing.
                    if isinstance(value, list):
                        return [str(fk.__pk__) for fk in value]
                    return str(value.__pk__)
                if field.metadata == AirModel.FIELD_LAZY and value == AirModel.FIELD_LAZY:
                    value = getattr(  # Resolve Foreign Key fields when their attributes are accessed.
                        self.get_model_api()(self.__api__).get(self.__pk__), name
                    )
                    setattr(self, name, value)
        return value

    def __refresh__(self, refreshed_obj: Optional[BaseModel] = None) -> None:
        """Refreshed the instances data from the backend.

        Raises
        ------
        NotImplementedError - When the model's API does not implement `get`.
        """
        if refreshed_obj is None:
            endpoint_api = self.get_model_api()(self.__api__)
            if endpoint_api is None:
                raise NotImplementedError
            refreshed_obj = endpoint_api.get(pk=self.__pk__)
        for field in fields(self):
            setattr(self, field.name, getattr(refreshed_obj, field.name))

    @classmethod
    @abstractmethod
    def get_model_api(
        cls: Type[TAirModel_co],
    ) -> Type[BaseEndpointApi[TAirModel_co]]:
        """
        Returns the respective `AirModelAPI` type for this model.
        """

    def refresh(self) -> None:
        """Refresh the instance by querying new API data.

        This uses the `get` method on the model's EndpointApi by default.
        """
        self.__refresh__()

    def _ensure_pk_exists(self, context: str) -> None:
        # We cannot perform detailed API calls for instances without populated primary keys
        if self.__pk__ is None:
            raise AirError(f'The {self.__class__.__name__} cannot be {context}: primary key is `None`.')

    def update(self, *args: Any, **kwargs: Any) -> None:
        self._ensure_pk_exists('updated')
        updated_obj = self.get_model_api()(self.__api__).patch(self.__pk__, **kwargs)
        self.__refresh__(updated_obj)  # Ensure update data is reflected in model instance.

    def full_update(self, *args: Any, **kwargs: Any) -> None:
        self._ensure_pk_exists('fully updated')
        updated_obj = self.get_model_api()(self.__api__).put(self.__pk__, **kwargs)
        self.__refresh__(updated_obj)  # Ensure update data is reflected in model instance.

    def delete(self) -> None:
        """Delete the instance and nullify the primary key."""
        self._ensure_pk_exists('deleted')
        self.get_model_api()(self.__api__).delete(self.__pk__)
        setattr(self, self.primary_key_field, None)


class ForeignKeyMixin(Generic[TAirModel_co]):
    """AirModel mixin for lazily resolving the instance."""

    def __init__(self, primary_key: UUID, _api: AirApi) -> None:
        self.__fk__ = primary_key
        self.__fk_resolved__ = False
        self.__api__ = _api

    @property
    def __pk__(self) -> PrimaryKey:
        return self.__fk__

    def __getattribute__(self, name: str) -> Any:
        """Loads the instance upon initial access to an exposed attribute."""
        if not is_dunder(name) and as_field(self, name) is not None and not self.__fk_resolved__:
            self.__refresh__()
            self.__fk_resolved__ = True

        return super().__getattribute__(name)


class ApiNotImplementedMixin:
    """Mixin used to allow AirModel subclasses to have an unimplemented API."""

    def __refresh__(self, refreshed_obj: Optional[BaseModel] = None) -> None:
        if refreshed_obj is None and getattr(self, '__fk_resolved__', True) is True:
            raise NotImplementedError
        super().__refresh__(refreshed_obj)  # type: ignore[misc]


class EndpointMethodMixin:
    """Mixin class for defining common endpoint methods.

    This is used to prevent with the intention of raising
    a `NotImplementedError` instead of an `AttributeError` when
    specific endpoint methods are not implemented in the SDK or API.
    """

    def list(self, **kwargs: Any) -> Iterator[AirModel]:
        raise NotImplementedError

    def create(self, *args: Any, **kwargs: Any) -> AirModel:
        raise NotImplementedError

    def get(self, *args: Any, **kwargs: Any) -> AirModel:
        raise NotImplementedError

    def put(self, *args: Any, **kwargs: Any) -> AirModel:
        raise NotImplementedError

    def patch(self, *args: Any, **kwargs: Any) -> AirModel:
        raise NotImplementedError

    def delete(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError


class BaseEndpointApi(EndpointMethodMixin, Generic[TAirModel_co]):
    model: Type[TAirModel_co]
    API_PATH: str = ''
    url: str

    def __init__(self, api: AirApi):
        self.__api__ = api
        self.url = join_urls(self.__api__.client.base_url, self.API_PATH)

    @cached_property
    def model_cls_type_hints(self) -> DataDict:
        return get_type_hints(self.model)

    @cached_property
    def model_cls_fields(self) -> Tuple[Field[Any], ...]:
        return fields(self.model)

    def load_model(self, data: DataDict) -> TAirModel_co:
        """Construct a new model instance, validate data, and set the API Client."""
        provided_fields = [f for f in self.model_cls_fields if f.name in data]
        missing_fields = [f for f in self.model_cls_fields if f.name not in data]

        try:
            model_inst = self.model(
                _api=self.__api__,
                **self.parse_provided_fields(provided_fields, data),
                **self.get_defaults_for_special_missing_fields(missing_fields),
            )
            return model_inst
        except TypeError as e:
            raise AirModelAttributeError(f'failed to instantiate `{self.model}`: {e}') from None

    def get_defaults_for_special_missing_fields(self, dc_fields: List[Field[Any]]) -> DataDict:
        special_fields: Dict[str, Union[object, None]] = {}
        for field in dc_fields:
            if field.metadata == AirModel.FIELD_LAZY:
                # lazy fields which are not present are assigned a placeholder value
                special_fields[field.name] = AirModel.FIELD_LAZY
            elif is_optional_union(self.model_cls_type_hints[field.name]):
                # optional fields which are not present are assigned to `None`
                special_fields[field.name] = None
        return special_fields

    def parse_provided_fields(self, dc_fields: List[Field[Any]], data: DataDict) -> DataDict:
        return {
            field.name: self.parse_field(
                self.model_cls_type_hints[field.name],
                field.metadata,
                data[field.name],
                f'field `{field.name}` of `{self.model.__name__}`',
            )
            for field in dc_fields
        }

    def parse_field(
        self,
        hint: Type[T],
        metadata: Mapping[Any, Any],
        provided_value: Any,
        context: str,
    ) -> T:
        """Parse the provided value based on the type hint of the value.

        This allows us to perform type checking of provided values and assists in
        the implementation of our `FIELD_FOREIGN_KEY` and `FIELD_LAZY` fields.
        """
        origin = get_origin(hint)
        try:
            if origin is not None:
                if isinstance(origin, type) and issubclass(origin, list):
                    return cast(
                        T, self.handle_list_field(cast(Type[List[T]], hint), metadata, provided_value)
                    )
                elif isinstance(origin, type) and issubclass(origin, dict):
                    return cast(T, provided_value)
                elif is_optional_union(hint):  # field is optional
                    return cast(
                        T, self.handle_optional_field(cast(Type[Optional[T]], hint), metadata, provided_value)
                    )
            elif isinstance(hint, type):  # field is an AirModel
                if issubclass(hint, AirModel):
                    return self.handle_air_model_field(hint, metadata, provided_value)  # type: ignore
                elif is_dataclass(hint) and isinstance(provided_value, dict):
                    return cast(hint, hint(**provided_value))  # type: ignore[valid-type]
                elif issubclass(hint, datetime):  # field is a datetime object
                    return cast(hint, self.handle_datetime_field(provided_value))  # type: ignore
                elif issubclass(hint, (int, str, bool, float)):  # field is a primitive
                    return self.handle_primitive_field(hint, provided_value)  # type: ignore
            if (
                is_union(hint)
                and type_check(provided_value, hint)
                and isinstance(provided_value, (bool, str, int, list, dict, float))
            ):
                return provided_value  # type: ignore
            raise AirModelAttributeError()
        except AirModelAttributeError as e:
            e.args = (f'{context}: {e}',) + e.args[1:]
            raise

    def handle_list_field(
        self, hint: Type[List[T]], metadata: Mapping[Any, Any], provided_value: Union[Any, List[Any]]
    ) -> List[T]:
        """
        Provided `data` argument is validated to be an actual list.
        Each item in `data` list is then validated against the target type and parsed individually.
        """
        if not isinstance(provided_value, list):
            raise AirModelAttributeError(
                f'field data is of type `{type(provided_value).__name__}`, expected `{list}`: {provided_value}'
            )

        return [
            self.parse_field(get_list_arg(hint), metadata, data_item, f'item at index `{index}`')
            for index, data_item in enumerate(provided_value)
        ]

    def handle_optional_field(
        self,
        hint: Type[Optional[T]],
        metadata: Mapping[Any, Any],
        provided_value: Any,
    ) -> Optional[T]:
        if provided_value is None:
            return None
        return self.parse_field(get_optional_arg(hint), metadata, provided_value, 'optional field')

    def handle_air_model_field(
        self,
        hint: Type[TAirModel_co],
        metadata: Mapping[Any, Any],
        provided_value: Any,
    ) -> TAirModel_co:
        """
        `AirModel` fields are validated as follows:
        - Foreign key fields are reserved for on-demand loading
        - Otherwise, field is parsed as a regular `BaseModel`
        """
        if metadata == hint.FIELD_FOREIGN_KEY:
            if isinstance(provided_value, str) or isinstance(provided_value, UUID):
                if primary_key := to_uuid(str(provided_value)):
                    lazy_fk = cast(
                        hint,  # type: ignore
                        type(str(hint.__name__), (ForeignKeyMixin, hint), {})(primary_key, _api=self.__api__),
                    )
                    return lazy_fk
                raise AirModelAttributeError(
                    f'`{hint.__name__}` can not be parsed from foreign key due to invalid UUID value: {provided_value}'
                )
            elif isinstance(provided_value, hint):
                return provided_value

        if isinstance(provided_value, dict):
            fk_api = hint.get_model_api()(self.__api__)
            return fk_api.load_model(provided_value)

        raise AirModelAttributeError(
            f'`{hint.__name__}` can not be parsed from foreign key due to invalid value: {provided_value}'
        )

    def handle_datetime_field(self, provided_value: Any) -> datetime:
        if isinstance(provided_value, datetime):
            return provided_value

        elif isinstance(provided_value, str):
            value = iso_string_to_datetime(provided_value)
            if value is None:
                raise AirModelAttributeError(f'field data is not a valid ISO string: {provided_value}')

            return value

        raise AirModelAttributeError(
            f'`{datetime}` field can not be parsed from field data of type `{type(provided_value).__name__}`: {provided_value}'
        )

    def handle_primitive_field(
        self,
        hint: Type[TSupportedPrimitive],
        provided_value: Any,
    ) -> TSupportedPrimitive:
        """Primitive fields are validated for type mismatch and returned as-is."""
        if not isinstance(provided_value, hint):
            raise AirModelAttributeError(
                f'field data is of type `{type(provided_value).__name__}`, expected `{hint.__name__}`: {provided_value}'
            )
        return provided_value



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT


from datetime import datetime
from json import JSONEncoder
from typing import Any
from uuid import UUID

from air_sdk.v2.utils import datetime_to_iso_string


class AirJSONEncoder(JSONEncoder):
    """`JSONEncoder` with Air-specific decoding logic."""

    def default(self, o: Any) -> Any:
        """`JSONEncoder.default` with ability to decode `datetime` and UUID objects."""
        from air_sdk.v2.air_model import AirModel

        if isinstance(o, datetime):
            return datetime_to_iso_string(o)
        if isinstance(o, UUID):
            return str(o)
        if isinstance(o, AirModel):
            pk = getattr(o, '__pk__', None)
            if pk is None:
                raise ValueError(
                    f'The `{o.__class__.__name__}` provided is not JSON serializable: __pk__ is None'
                )
            return str(pk)
        return super().default(o)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.simulations import Simulation
from air_sdk.v2.endpoints.workers import Worker
from air_sdk.v2.air_model import AirModel, BaseEndpointApi
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class Job(AirModel):
    id: str
    category: str
    state: str
    created: datetime = field(repr=False)
    last_updated: datetime = field(repr=False)
    notes: Optional[str] = field(repr=False)
    data: Optional[str] = field(repr=False)
    simulation: Optional[Simulation] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    worker: Optional[Worker] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return JobEndpointApi

    @validate_payload_types
    def update(self, state: str) -> None:
        super().update(state=state)


class JobEndpointApi(
    mixins.ListApiMixin[Job],
    mixins.GetApiMixin[Job],
    mixins.PatchApiMixin[Job],
    BaseEndpointApi[Job],
):
    API_PATH = 'jobs'
    model = Job



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Helper utils
"""

import datetime
from json import JSONDecodeError
from urllib.parse import ParseResult
from requests import Response

from dateutil import parser as dateparser

from .exceptions import AirUnexpectedResponse
from .logger import air_sdk_logger as logger


def raise_if_invalid_response(res: Response, status_code=200, data_type=dict):
    """
    Validates that a given API response has the expected status code and JSON payload

    Arguments:
    res (requests.HTTPResponse) - API response object
    status_code [int] - Expected status code (default: 200)

    Raises:
    AirUnexpectedResponse - Raised if an unexpected response is received from the API
    """
    json = None
    if res.status_code != status_code:
        logger.debug(res.text)
        raise AirUnexpectedResponse(message=res.text, status_code=res.status_code)
    if not data_type:
        return
    try:
        json = res.json()
    except JSONDecodeError:
        raise AirUnexpectedResponse(message=res.text, status_code=res.status_code)
    if not isinstance(json, data_type):
        raise AirUnexpectedResponse(
            message=f'Expected API response to be of type {data_type}, ' + f'got {type(json)}',
            status_code=res.status_code,
        )


def required_kwargs(required):
    """Decorator to enforce required kwargs for a function"""
    if not isinstance(required, list):
        required = [required]

    def wrapper(method):
        def wrapped(*args, **kwargs):
            for arg in required:
                if isinstance(arg, tuple):
                    present = False
                    for option in arg:
                        if option in kwargs:
                            present = True
                            break
                    if not present:
                        raise AttributeError(f'{method} requires one of the following: {arg}')
                else:
                    if arg not in kwargs:
                        raise AttributeError(f'{method} requires {arg}')
            return method(*args, **kwargs)

        return wrapped

    return wrapper


def deprecated(new=None):
    """Decorator to log a warning when calling a deprecated function"""

    def wrapper(method):
        def wrapped(*args, **kwargs):
            msg = f'{method} has been deprecated and will be removed in a future release.'
            if new:
                msg += f' Use {new} instead.'
            logger.warning(msg)
            return method(*args, **kwargs)

        return wrapped

    return wrapper


def validate_timestamps(log_prefix, **kwargs):
    """
    Logs a warning if any provided timestamps are in the past

    Arguments:
        log_prefix (str): Prefix to be prepended to the logged warning(s)
        kwargs (dict): Timestamps to verify
    """
    now = datetime.datetime.now()
    for key, value in kwargs.items():
        if value and dateparser.parse(str(value)) <= now:
            logger.warning(f'{log_prefix} with `{key}` in the past: {value} (now: {now})')


def is_datetime_str(value):
    """
    Checks to see if the string is a valid datetime format

    Arguments:
        value (str): String to test if valid datetime format
    """
    if isinstance(value, str):
        try:
            return datetime.datetime.fromisoformat(value.replace('Z', '+00:00'))
        except ValueError:
            pass
    return False


def url_path_join(base: ParseResult, *segments: str, trailing_slash: bool = False) -> ParseResult:
    """
    Appends provided path segments (if any) to provided base URL.
    Appends or removes a trailing slash at the end of path as specified by `trailing_slash` argument.

    Examples:
    ```
    url = urlparse("https://example.com/a/b")

    url_path_join(url, "c", "d")                        # https://example.com/a/b/c/d
    url_path_join(url, "c", "d", trailing_slash=True)   # https://example.com/a/b/c/d/
    url_path_join(url, trailing_slash=True)             # https://example.com/a/b/
    ```
    """

    resulting_path = base.path
    for segment in segments:
        resulting_path = f"{resulting_path.rstrip('/')}/{segment.lstrip('/')}"

    resulting_url = base._replace(path=resulting_path.rstrip('/') + ('/' if trailing_slash else ''))
    return resulting_url



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Constants shared throughout the SDK.
"""

ALLOWED_HOSTS = [
    'air.nvidia.com',
    'staging.air.nvidia.com',
    'air.cumulusnetworks.com',
    'staging.air.cumulusnetworks.com',
]

DEFAULT_API_URL = 'https://air.nvidia.com/api/'


DEFAULT_CONNECT_TIMEOUT = 16  # seconds
DEFAULT_READ_TIMEOUT = 61  # seconds
DEFAULT_PAGINATION_PAGE_SIZE = 200  # Objects per paginated response



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from typing import Any, Optional

from air_sdk.v2.air_model import AirModel, BaseEndpointApi, PrimaryKey, ApiNotImplementedMixin


@dataclass(eq=False)
class ResourceBudget(ApiNotImplementedMixin, AirModel):
    id: str
    cpu: Optional[int] = field(repr=False)
    cpu_used: Optional[int] = field(repr=False)
    image_uploads: Optional[int] = field(repr=False)
    image_uploads_used: Optional[int] = field(repr=False)
    memory: Optional[int] = field(repr=False)
    memory_used: Optional[int] = field(repr=False)
    simulations: Optional[int] = field(repr=False)
    simulations_used: Optional[int] = field(repr=False)
    storage: Optional[int] = field(repr=False)
    storage_used: Optional[int] = field(repr=False)
    userconfigs: Optional[int] = field(repr=False)
    userconfigs_used: Optional[int] = field(repr=False)

    @classmethod
    def get_model_api(cls):
        return ResourceBudgetEndpointApi


class ResourceBudgetEndpointApi(BaseEndpointApi[ResourceBudget]):
    API_PATH = 'resource_budgets'  # A placeholder
    model = ResourceBudget

    def get(self, pk: PrimaryKey, **params: Any) -> ResourceBudget:
        return self.load_model({'id': str(pk)})



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT


from dataclasses import field, dataclass
from datetime import datetime
from typing import Optional, Union

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.interfaces import Interface
from air_sdk.v2.endpoints.simulations import Simulation
from air_sdk.v2.air_model import AirModel, BaseEndpointApi, PrimaryKey
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class Service(AirModel):
    id: str
    name: str
    created: datetime
    modified: datetime
    dest_port: int
    src_port: int
    service_type: str
    interface: Interface = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    simulation: Simulation = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    host: Optional[str]

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return ServiceEndpointApi


class ServiceEndpointApi(
    mixins.ListApiMixin[Service],
    mixins.CreateApiMixin[Service],
    mixins.DeleteApiMixin,
    BaseEndpointApi[Service],
):
    API_PATH = 'simulations/nodes/interfaces/services/'
    model = Service

    @validate_payload_types
    def create(
        self, name: str, dest_port: int, interface: Union[Interface, PrimaryKey], service_type: str = 'ssh'
    ) -> Service:
        return super().create(
            name=name,
            dest_port=dest_port,
            interface=interface,
            service_type=service_type,
        )



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.resource_budgets import ResourceBudget
from air_sdk.v2.air_model import AirModel, BaseEndpointApi


@dataclass(eq=False)
class Organization(AirModel):
    id: str
    name: str
    member_count: int
    resource_budget: ResourceBudget = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return OrganizationEndpointApi


class OrganizationEndpointApi(
    mixins.ListApiMixin[Organization],
    mixins.GetApiMixin[Organization],
    BaseEndpointApi[Organization],
):
    API_PATH = 'organizations'
    model = Organization



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
from http import HTTPStatus
from typing import Optional, Union, TYPE_CHECKING, List, TypedDict, cast, Any, Dict

from air_sdk.util import raise_if_invalid_response
from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.images import Image
from air_sdk.v2.endpoints.mixins import serialize_payload
from air_sdk.v2.endpoints.simulations import Simulation
from air_sdk.v2.endpoints.systems import System
from air_sdk.v2.endpoints.workers import Worker
from air_sdk.v2.air_model import AirModel, BaseEndpointApi, PrimaryKey, DataDict

from air_sdk.v2.utils import validate_payload_types, join_urls


if TYPE_CHECKING:
    from air_sdk.v2.endpoints.cloud_inits import CloudInit


class BulkUpdateSimNodeStateType(TypedDict):
    """Type hints for the payload of the `NodeEndpointApi.bulk_update_state` payload."""

    state: str
    ids: List[PrimaryKey]


@dataclass(eq=False)
class Node(AirModel):
    id: str
    name: str
    created: datetime = field(repr=False)
    modified: datetime = field(repr=False)
    simulation: Simulation = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    os: Optional[Image] = field(metadata=AirModel.FIELD_FOREIGN_KEY)
    console_port: Optional[int] = field(repr=False)
    serial_port: Optional[int] = field(repr=False)
    state: str
    memory: int = field(repr=False)
    storage: int = field(repr=False)
    cpu: int = field(repr=False)
    version: int = field(repr=False)
    features: str = field(repr=False)
    simx_ipv4: Optional[str] = field(repr=False)  # We should remove this when plugin API is released
    worker: Optional[Worker] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    system: Optional[System] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    boot_group: Optional[int] = field(repr=False)
    console_url: Optional[str] = field(repr=False)
    console_username: Optional[str] = field(repr=False)
    console_password: Optional[str] = field(repr=False)
    metadata: Optional[str] = field(repr=False)
    pos_x: float = field(repr=False)
    pos_y: float = field(repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return NodeEndpointApi

    @validate_payload_types
    def update(
        self,
        name: Optional[str] = None,
        state: Optional[str] = None,
        pos_x: Optional[float] = None,
        pos_y: Optional[float] = None,
        boot_group: Optional[int] = None,
        console_username: Optional[str] = None,
        console_password: Optional[str] = None,
        console_port: Optional[int] = None,
        serial_port: Optional[int] = None,
    ) -> None:
        data = {
            'name': name,
            'pos_x': pos_x,
            'pos_y': pos_y,
            'boot_group': boot_group,
            'console_username': console_username,
            'console_password': console_password,
            'console_port': console_port,
            'serial_port': serial_port,
            'state': state,
        }
        data = {key: value for (key, value) in data.items() if value is not None}
        super().update(**data)

    @validate_payload_types
    def full_update(
        self,
        name: str,
        state: str,
        pos_x: float,
        pos_y: float,
        boot_group: Optional[int],
        console_username: Optional[str],
        console_password: Optional[str],
        console_port: Optional[int],
        serial_port: Optional[int],
    ) -> None:
        super().update(
            name=name,
            state=state,
            pos_x=pos_x,
            pos_y=pos_y,
            boot_group=boot_group,
            console_username=console_username,
            console_password=console_password,
            console_port=console_port,
            serial_port=serial_port,
        )

    @validate_payload_types
    def set_agent_key(self, agent_key: Optional[str]) -> None:
        """Set the Air Agent key for the node.

        This is only used by worker clients when starting simulation nodes.
        """
        super().update(agent_key=agent_key)

    @property
    def cloud_init(self) -> CloudInit:
        cloud_int: CloudInit = self.__api__.cloud_inits.get(self.__pk__)
        return cloud_int


class NodeEndpointApi(
    mixins.ListApiMixin[Node],
    mixins.CreateApiMixin[Node],
    mixins.GetApiMixin[Node],
    mixins.PatchApiMixin[Node],
    mixins.DeleteApiMixin,
    BaseEndpointApi[Node],
):
    API_PATH = 'simulations/nodes/'
    model = Node

    @validate_payload_types
    def create(
        self,
        simulation: Union[Simulation, PrimaryKey],
        name: str,
        pos_x: float = 0.0,
        pos_y: float = 0.0,
        boot_group: Optional[int] = 50,
        features: str = '{}',
        metadata: str = '{}',
        os: Optional[Union[Image, PrimaryKey]] = None,
        system: Optional[Union[System, PrimaryKey]] = None,
        memory: Optional[int] = None,
        cpu: Optional[int] = None,
        storage: Optional[int] = None,
        console_username: Optional[str] = None,
        console_password: Optional[str] = None,
    ) -> Node:
        payload: DataDict = {
            'simulation': simulation,
            'name': name,
            'os': os,
            'boot_group': boot_group,
            'features': features,
            'pos_x': pos_x,
            'pos_y': pos_y,
            'console_username': console_username,
            'console_password': console_password,
            'metadata': metadata,
        }
        for key, value in (
            ('system', system),
            ('storage', storage),
            ('memory', memory),
            ('cpu', cpu),
        ):
            if value is not None:
                payload[key] = value
        return super().create(**payload)

    @validate_payload_types
    def bulk_update_state(self, data: List[BulkUpdateSimNodeStateType]) -> None:
        """Method used by worker clients to perform state updates of nodes in bulk."""
        response = self.__api__.client.patch(
            join_urls(self.url, 'bulk-update-state'),
            data=serialize_payload(cast(List[Dict[str, Any]], data)),
        )
        raise_if_invalid_response(response, status_code=HTTPStatus.OK, data_type=None)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Any, Union

from air_sdk.util import raise_if_invalid_response
from air_sdk.v2.air_model import AirModel, PrimaryKey, BaseEndpointApi
from air_sdk.v2.endpoints.mixins import serialize_payload
from air_sdk.v2.endpoints.nodes import Node
from air_sdk.v2.endpoints.user_configs import UserConfig
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class CloudInit(AirModel):
    simulation_node: Node = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    user_data: Optional[UserConfig] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    meta_data: Optional[UserConfig] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    user_data_name: Optional[str] = field(repr=False)
    meta_data_name: Optional[str] = field(repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return CloudInitEndpointApi

    @property
    def primary_key_field(self) -> str:
        return 'simulation_node'

    @property
    def __pk__(self) -> PrimaryKey:
        return getattr(self, self.primary_key_field).__pk__  # type: ignore

    @validate_payload_types
    def full_update(
        self,
        user_data: Optional[Union[UserConfig, PrimaryKey]],
        meta_data: Optional[Union[UserConfig, PrimaryKey]],
    ) -> None:
        """Update all fields of the cloud-init assignment."""
        super().update(user_data=user_data, meta_data=meta_data)


class CloudInitEndpointApi(BaseEndpointApi[CloudInit]):
    API_PATH = 'simulations/nodes/{id}/cloud-init'  # Placeholder
    model = CloudInit

    def get(self, pk: PrimaryKey, **params: Any) -> CloudInit:
        detail_url = self.url.format(id=str(pk))
        response = self.__api__.client.get(detail_url, params=params)
        raise_if_invalid_response(response)
        return self.load_model(response.json())

    def patch(self, pk: PrimaryKey, **kwargs: Any) -> CloudInit:
        detail_url = self.url.format(id=str(pk))
        response = self.__api__.client.patch(detail_url, data=serialize_payload(kwargs))
        raise_if_invalid_response(response)
        return self.load_model(response.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Organization module
"""

from . import util
from .air_model import AirModel
from .fleet import Fleet, FleetApi


class Organization(AirModel):
    """
    Manage an Organization

    ### delete
    Delete the organization. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the organization

    ### refresh
    Syncs the organization with all values returned by the API

    ### update
    Update the organization with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    ORG_MEMBER_ROLE = 'Organization Member'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._members_api_url = f'{self._api.url}{self.id}/members/'

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Organization {self.name} {self.id}>'

    def add_member(self, username: str, roles: list = None):
        """
        Add a new member to the organization

        Arguments:
            username (str): The email address of the user to add
            roles (list, optional): A list of roles to assign the user. Valid values are
                'Organization Admin' or 'Organization Member'. If no roles list is provided,
                'Organization Member' is used as the default role.

        Example:
        ```
        >>> organization.add_member('user1@nvidia.com')
        >>> organization.add_member('user2@nvidia.com', roles=['Organization Admin'])
        ```
        """
        _roles = roles
        if not _roles:
            _roles = [self.ORG_MEMBER_ROLE]
        self._api.client.post(self._members_api_url, json={'username': username, 'roles': _roles})
        self.refresh()

    def add_members(self, members: list):
        # pylint: disable=line-too-long
        """
        Add new members to the organization

        Arguments:
            members (list): List of organization membership dicts in the format of
                {'username': <email_address>, 'roles': [<role>]}.
                'roles' is optional and defaults to ['Organization Member']
                <role> can be a value of 'Organization Admin' or 'Organization Member'.

        Example:
        ```
        >>> organization.add_members([{'username': 'user1@nvidia.com', 'roles': ['Organization Admin']}, {'username': 'user2@nvidia.com'}])
        ```
        """  # pylint: enable=line-too-long
        for member in members:
            if not member.get('roles', []):
                member['roles'] = [self.ORG_MEMBER_ROLE]
        self._api.client.post(self._members_api_url, json=members)
        self.refresh()

    def remove_member(self, username: str, **kwargs):
        """
        Remove a member from the organization

        Arguments:
            username (str): The email address of the user to remove

        Example:
        ```
        >>> organization.remove_member('user1@nvidia.com')
        """
        self._api.client.delete(self._members_api_url, json={'username': username})
        if kwargs.get('_refresh_when_done', True):
            self.refresh()

    def remove_members(self, members: list):
        """
        Remove multiple members from the organization

        Arguments:
            members (list): Email addresses of the users to remove

        Example:
        ```
        >>> organization.remove_members(['user1@nvidia.com', 'user2@nvidia.com'])
        """
        for member in members:
            self.remove_member(member, _refresh_when_done=False)
        self.refresh()

    @util.required_kwargs(['name'])
    def create_fleet(self, **kwargs):
        # pylint: disable=line-too-long
        """
        Create a new fleet for this organization
        Arguments:
            name (str): Name of the fleet
            prefix (str, optional): Prefix for the fleet's IPv6 addresses
            prefix_length (int, optional): Length of the IPv6 address prefix
            gateway_ipv4 (str, optional): IPv4 gateway address for the fleet
            port_range (str, optional): Range of ports available on the fleet
            container_ipv4_network (str, optional): IPv4 network address for containers in the fleet
            container_prefix (int, optional): Prefix length for container IP addresses
            labels (list, optional): List of labels to assign to the fleet
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
            [`Fleet`](/docs/fleet)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organization.create_fleet(name='MyFleet', prefix='fd10:1:1::', prefix_length=48, gateway_ipv4='10.0.0.1', port_range='20000-40000', container_ipv4_network='172.16.0.0', container_prefix=16, labels=['label1', 'label2'])
        <Fleet MyFleet 3dadd54d-583c-432e-9383-a2b0b1d7f221>
        ```
        """  # pylint: enable=line-too-long
        url = FleetApi(self._api.client).url
        kwargs['organization'] = self.id
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Fleet(self, **res.json())


class OrganizationApi:
    """High-level interface for the Organization API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/organization/'

    def get(self, organization_id, **kwargs):
        """
        Get an existing organization

        Arguments:
            organization_id (str): Organization ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Organization`](/docs/organization)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organizations.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Organization NVIDIA 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{organization_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Organization(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing organizations

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organizations.list()
        [<Organization NVIDIA c51b49b6-94a7-4c93-950c-e7fa4883591>, <Organization Customer 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Organization(self, **organization) for organization in res.json()]

    @util.required_kwargs(['name'])
    def create(self, **kwargs):
        # pylint: disable=line-too-long
        """
        Create a new organization

        Arguments:
            name (str): Organization name
            members (list, optional): List of organization membership dicts in the format of
                {'username': <email_address>, 'roles': [<role>]}.
                'roles' is optional and defaults to ['Organization Member']
                <role> can be a value of 'Organization Admin' or 'Organization Member'.
                If no member list is provided, the calling user's account will be set as the
                organization admin by default.
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Organization`](/docs/organization)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organizations.create(name='NVIDIA', members=[{'username': 'user1@nvidia.com', 'roles': ['Organization Admin']}, {'username': 'user2@nvidia.com'}])
        <Organization NVIDIA 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """  # pylint: enable=line-too-long
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Organization(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any

from air_sdk.v2.air_model import AirModel, BaseEndpointApi, PrimaryKey, ApiNotImplementedMixin


@dataclass(eq=False)
class System(ApiNotImplementedMixin, AirModel):
    id: str

    @classmethod
    def get_model_api(cls):
        return SystemEndpointApi


class SystemEndpointApi(BaseEndpointApi[System]):
    API_PATH = 'systems'  # A placeholder
    model = System

    def get(self, pk: PrimaryKey, **params: Any) -> System:
        return self.load_model({'id': str(pk)})



# SPDX-FileCopyrightText: Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
ResourceBudget module
"""

from . import util
from .air_model import AirModel


class ResourceBudget(AirModel):
    """
    Manage a ResourceBudget

    ### json
    Returns a JSON string representation of the budget

    ### refresh
    Syncs the budget with all values returned by the API

    ### update
    Update the budget with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    _deletable = False

    def __repr__(self):
        if self._deleted:
            return super().__repr__()
        return f'<ResourceBudget {self.id}>'


class ResourceBudgetApi:
    """High-level interface for the ResourceBudget API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/resource-budget/'

    def get(self, budget_id, **kwargs):
        """
        Get an existing budget

        Arguments:
            budget_id (str): ResourceBudget ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`ResourceBudget`](/docs/resourcebudget)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.resource_budgets.get('c604c262-396a-48a0-a8f6-31708c0cff82')
        <ResourceBudget c604c262-396a-48a0-a8f6-31708c0cff82>
        ```
        """
        url = f'{self.url}{budget_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return ResourceBudget(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing budgets

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.resource_budgets.list()
        [<ResourceBudget c604c262-396a-48a0-a8f6-31708c0cff82>, <ResourceBudget 906675f7-8b8d-4f52-b59d-52847af2f0ef>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [ResourceBudget(self, **budget) for budget in res.json()]



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Link module
"""

from . import util
from .air_model import AirModel


class Link(AirModel):
    """
    Manage a Link

    ### delete
    Delete the link. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the link

    ### refresh
    Syncs the link with all values returned by the API

    ### update
    Update the link with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    def __repr__(self):
        if self._deleted:
            return super().__repr__()
        return f'<Link {self.id}>'


class LinkApi:
    """High-level interface for the Link API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/link/'

    def get(self, link_id, **kwargs):
        """
        Get an existing link

        Arguments:
            link_id (str): Link ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Link`](/docs/link)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.links.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Link 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{link_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Link(self, **res.json())

    def list(self, **kwargs):
        """
        List existing links

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.links.list()
        [<Link c51b49b6-94a7-4c93-950c-e7fa4883591>, <Link 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Link(self, **link) for link in res.json()]

    @util.required_kwargs(['topology', 'interfaces'])
    def create(self, **kwargs):
        # pylint: disable=line-too-long
        """
        Create a new link

        Arguments:
            topology (str | `Topology`): `Topology` or ID
            interfaces (list): List of `Interface` objects or IDs
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Link`](/docs/link)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.links.create(topology=topology, interfaces=[intf1, 'fd61e3d8-af2f-4735-8b1d-356ee6bf4abe'])
        <Link 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        ```
        """  # pylint: enable=line-too-long
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Link(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT
import inspect
from typing import Any, List, Literal, Optional, Tuple, Type, TypeVar, Union, get_args, get_origin
import sys
from dataclasses import is_dataclass
from typing import (
    ForwardRef,
)

T = TypeVar('T')


def union_args_are_optional(args: Tuple[Union[Any, Any], ...]) -> bool:
    return len(args) >= 2 and type(None) in args


def is_union(type_: Type[Any]) -> bool:
    return get_origin(type_) == Union


def is_optional_union(type_: Type[Any]) -> bool:
    return is_union(type_) and union_args_are_optional(get_args(type_))


def get_optional_arg(optional_type: Type[Optional[T]]) -> Type[T]:
    return next(arg for arg in get_args(optional_type) if arg is not type(None))  # type: ignore[no-any-return]


def get_list_arg(list_type: Type[List[T]]) -> Type[T]:
    return get_args(list_type)[0]  # type: ignore[no-any-return]


def is_typed_dict(expected_type: Type[Any]) -> bool:
    """Determine if the `expected_type` provided is a subclass of TypedDict."""
    return hasattr(expected_type, '__annotations__') and not is_dataclass(expected_type)


def type_check_typed_dict(value: Any, expected_type: Type[Any]) -> bool:
    """Perform type checking when the expected_type is a subclass of TypedDict.

    This currently does not work if the expected_type is also a dataclass.
    """
    if not isinstance(value, dict):
        return False
    expected_keys = expected_type.__annotations__.keys()
    # Check all keys provided are defined within the expected_type TypedDict
    if not all(key in value for key in expected_keys):
        return False
    # Recursively check each key's value type
    return all(type_check(value[key], expected_type.__annotations__[key]) for key in expected_keys)


def type_check(value: Any, expected_type: Type[Any]) -> bool:
    """Recursively check if the value matches the expected type."""
    from air_sdk.v2.air_model import PrimaryKey  # noqa

    if isinstance(expected_type, ForwardRef):
        if sys.version_info >= (3, 9):  # Python 3.9+ requires a third `recursive_guard` arg
            expected_type = expected_type._evaluate(globals(), locals(), frozenset())
        else:
            expected_type = expected_type._evaluate(globals(), locals())
    origin = get_origin(expected_type)
    args = get_args(expected_type)

    if origin is None:  # Base case
        if is_typed_dict(expected_type):
            return type_check_typed_dict(value, expected_type)
        return isinstance(value, expected_type)

    if origin is Union:
        return any(type_check(value, arg) for arg in args)

    if origin is list:
        if not isinstance(value, list):
            return False
        if not args:  # We're already a list, so if not args then we're good
            return True
        return all(type_check(item, args[0]) for item in value)

    if origin is dict:
        if not isinstance(value, dict):
            return False
        if not args:  # We're already a dict, so if no args then we're good
            return True
        key_type, value_type = args
        return all(type_check(k, key_type) and type_check(v, value_type) for k, v in value.items())

    if origin is Literal:
        return any(value == arg for arg in args)

    if inspect.isclass(origin):
        return isinstance(value, origin)

    return False



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Service module
"""

from . import util
from .air_model import AirModel


class Service(AirModel):
    """
    Manage a Service

    ### delete
    Delete the service. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the service

    ### refresh
    Syncs the service with all values returned by the API

    ### update
    Update the service with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Service {self.name} {self.id}>'


class ServiceApi:
    """High-level interface for the Service API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/service/'

    @util.deprecated('ServiceApi.list()')
    def get_services(self):  # pylint: disable=missing-function-docstring
        return self.list()

    @util.deprecated('ServiceApi.get()')
    def get_service(self, service_id):  # pylint: disable=missing-function-docstring
        return self.get(service_id)

    @util.deprecated('ServiceApi.create()')
    def create_service(self, simulation_id, name, interface, dest_port, **kwargs):  # pylint: disable=missing-function-docstring
        return self.create(
            simulation=simulation_id, name=name, interface=interface, dest_port=dest_port, **kwargs
        )

    def get(self, service_id, **kwargs):
        """
        Get an existing service

        Arguments:
            service_id (str): Service ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Service`](/docs/service)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.services.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Service SSH 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{service_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Service(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing services

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.services.list()
        [<Service SSH c51b49b6-94a7-4c93-950c-e7fa4883591>, <Service HTTP 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Service(self, **service) for service in res.json()]

    @util.required_kwargs(['name', 'simulation', 'interface'])
    def create(self, **kwargs):
        """
        Create a new service

        Arguments:
            name (str): Service name
            interface (str | `SimulationInterface`): Interface that the service should be created
                for. This can be provided in one of the following formats:
                - [`SimulationInterface`](/docs/simulationinterface) object
                - ID of a [`SimulationInterface`](/docs/simulationinterface)
                - String in the format of 'node_name:interface_name'
            simulation (str | `Simulation`): `Simulation` or ID
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Service`](/docs/service)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.services.create(name='myservice', interface='oob-mgmt-server:eth0', dest_port=22)
        <Service myservice cc18d746-4cf0-4dd3-80c0-e7df68bbb782>
        >>> air.services.create(name='myservice', interface=simulation_interface, dest_port=22)
        <Service myservice 9603d0d5-5526-4a0f-91b8-a600010d0091>
        ```
        """
        if isinstance(kwargs['interface'], str) and ':' in kwargs['interface']:
            kwargs['interface'] = self._resolve_interface(kwargs['interface'], kwargs['simulation'])
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Service(self, **res.json())

    def _resolve_interface(self, interface, simulation):
        try:
            node_name = interface.split(':')[0]
            interface_name = interface.split(':')[1]
        except (SyntaxError, IndexError):
            raise ValueError(
                '`interface` must be an Interface object or in the format of ' + '"node_name:interface_name"'
            )

        resolved = None
        for node in self.client.nodes.list(simulation=simulation):
            if node.name == node_name:
                for intf in node.interfaces:
                    if intf.name == interface_name:
                        resolved = intf
                        break
        if not resolved:
            raise ValueError('Interface ' + interface + ' does not exist')
        return self.client.simulation_interfaces.list(original=resolved, simulation=simulation)[0]



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Base classes for AIR object models
"""

from __future__ import annotations

import json
import re
from abc import ABC
from datetime import date, datetime
from http import HTTPStatus
from typing import TYPE_CHECKING, Dict, Generic, List, Optional, Type, TypeVar, Union, get_args
from urllib.parse import urlparse


from . import util
from .exceptions import AirObjectDeleted

# `AirApi` exposes resource APIs via its properties
# this means that we can not import `AirApi` for type hinting purposes as this will cause a circular import
# therefore, we only import `AirApi` during type checking
if TYPE_CHECKING:  # pragma: no cover
    from typing import Literal

    from .air_api import AirApi

TAirModel = TypeVar('TAirModel', bound='AirModel')


class AirModel:
    """Base class for AIR object models"""

    model_keys = {
        'account': 'accounts',
        'base_simulation': 'simulations',
        'bios': 'images',
        'connection': 'links',
        'creator': 'accounts',
        'demo': 'demos',
        'interface': 'simulation_interfaces',
        'interfaces': {
            'Node': 'interfaces',
            'SimulationNode': 'simulation_interfaces',
            'Link': 'interfaces',
        },
        'job': 'jobs',
        'last_worker': 'worker',
        'node': {
            'Interface': 'nodes',
            'NodeInstruction': 'simulation_nodes',
            'SimulationInterface': 'simulation_nodes',
            'TopologyInstruction': 'nodes',
        },
        'nodes': 'simulation_nodes',
        'original': {
            'SimulationInterface': 'interfaces',
            'SimulationNode': 'nodes',
        },
        'organization': 'organizations',
        'os': 'images',
        'preferred_worker': 'workers',
        'services': 'services',
        'simulation': 'simulations',
        'topology': 'topologies',
        'worker': 'workers',
        'fleet': 'fleets',
        'userconfig': 'userconfigs',
    }

    def __init__(self, api: AirModelAPI, **kwargs):
        self._deleted = False
        super().__setattr__('_updatable', getattr(self, '_updatable', True))
        super().__setattr__('_deletable', getattr(self, '_deletable', True))
        self._api = api
        self._load(**kwargs)

    def _load(self, **kwargs):
        for key, value in kwargs.items():
            _value = value
            datetime_obj = util.is_datetime_str(value)
            if datetime_obj:
                _value = datetime_obj
            if key in self.model_keys and value:
                if isinstance(value, list) and not isinstance(value, LazyLoadedList):
                    _value = LazyLoadedList(
                        [LazyLoaded(id=_get_item_id(item), model=self._get_model_key(key)) for item in value],
                        self._api,
                    )
                elif isinstance(value, (LazyLoaded, LazyLoadedList)):
                    _value = value
                elif value.startswith('http'):
                    _value = LazyLoaded(id=_value.split('/')[6], model=self._get_model_key(key))
                else:
                    _value = LazyLoaded(id=_value, model=self._get_model_key(key))
            super().__setattr__(key, _value)

    def __repr__(self):
        repr_str = super().__repr__()
        if self._deleted:
            repr_str = f'<Deleted Object ({repr_str})>'
        return repr_str

    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if name == '_deleted':
            return value
        if self._deleted:
            raise AirObjectDeleted(type(self))
        if isinstance(value, LazyLoaded):
            value = getattr(self._api.client, value.model).get(value.id)
            super().__setattr__(name, value)
        return value

    def __setattr__(self, name, value):
        if name == '_deleted' or not self._updatable:
            return super().__setattr__(name, value)
        try:
            original = super().__getattribute__(name)
            api = super().__getattribute__('_api')
            id = super().__getattribute__('id')  # pylint: disable=redefined-builtin
        except AttributeError:
            original = None
            api = None
            id = None
        if not name.startswith('_') and api and id and original != value:
            self._patch(name, value)
        return super().__setattr__(name, value)

    def _get_model_key(self, key):
        value = self.model_keys[key]
        if isinstance(value, dict):
            value = self.model_keys[key][self.__class__.__name__]
        return value

    def _patch(self, key, value):
        url = f'{self._api.url}{self.id}/'
        res = self._api.client.patch(url, json={key: value})
        util.raise_if_invalid_response(res)

    def update(self, **kwargs):
        """
        Update the object with the provided data

        Arguments:
            kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
        """
        if not self._updatable:
            raise NotImplementedError(f'{self.__class__.__name__} does not support updates')
        url = f'{self._api.url}{self.id}/'
        ignored_fields = getattr(self, '_ignored_update_fields', None)
        args = kwargs
        if ignored_fields:
            allowed_payload = {}
            for key, value in kwargs.items():
                if key not in ignored_fields:
                    allowed_payload[key] = value
            args = allowed_payload
        res = self._api.client.patch(url, json=args)
        util.raise_if_invalid_response(res)

    def delete(self, **kwargs):
        """
        Delete the object. Once successful, the object should no longer be used and will raise
        [`AirDeletedObject`](/docs/exceptions) when referenced.

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed
        """
        if not self._deletable:
            raise NotImplementedError(f'{self.__class__.__name__} does not support deletes')
        url = f'{self._api.url}{self.id}/'
        res = self._api.client.delete(url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)
        self._deleted = True

    def refresh(self):
        """Syncs the object with all values returned by the API"""
        self._load(**self._api.get(self.id).__dict__)

    def json(self):
        """Returns a JSON string representation of the object"""
        payload = {}
        for key, value in self.__dict__.items():
            if isinstance(value, (datetime, date)):
                value = value.isoformat()
            if key.startswith('_'):
                continue
            if callable(value):
                continue
            if isinstance(value, (AirModel, LazyLoaded)):
                payload[key] = value.id
            elif isinstance(value, LazyLoadedList):
                payload[key] = [obj.id for obj in value.__iter__(skip_load=True)]
            else:
                payload[key] = value
        return json.dumps(payload)


class LazyLoaded:
    """A lazy object whose data will be loaded later"""

    def __init__(self, id, model):  # pylint: disable=redefined-builtin
        self.id = id
        self.model = model

    def __repr__(self):
        model_str = self.model.capitalize()
        if model_str == 'Topologies':
            model_str = 'Topology'
        elif model_str.endswith('s'):
            model_str = model_str[:-1]
        return f'<air_sdk.air_model.LazyLoaded {model_str} {self.id}>'


class LazyLoadedList(list):
    """A list whose items are LazyLoaded"""

    def __init__(self, items, api):
        self._api = api
        super().__init__(items)

    def __getitem__(self, index):
        value = super().__getitem__(index)
        if isinstance(value, LazyLoaded):
            value = getattr(self._api.client, value.model).get(value.id)
            self[index] = value
        return value

    def __iter__(self, skip_load=False):
        items = super().__iter__()
        for item in items:
            if isinstance(item, LazyLoaded) and not skip_load:
                yield getattr(self._api.client, item.model).get(item.id)
            else:
                yield item


def _get_item_id(item):
    if isinstance(item, dict):
        return item['id']
    try:
        return item.split('/')[6]
    except (AttributeError, IndexError):
        return item


class AirModelAPI(ABC, Generic[TAirModel]):
    """
    Generic class for representing Air resource APIs. Implements functionalities which are common across all resources.

    Usage example:
    ```
    class UserConfigAPI(AirModelAPI[UserConfig]):
        \"\"\"High-level interface for the UserConfig API.\"\"\"

        API_VERSION = 2
        API_PATH = 'userconfigs'
    ```
    """

    API_VERSION: int = 1
    API_PATH: Optional[str] = None

    def __init__(self, client: AirApi):
        if not self.API_PATH:
            raise AttributeError('Model API path `API_PATH` is not properly defined')

        self.client = client
        parsed_url = urlparse(self.client.api_url, allow_fragments=False)
        parsed_url = parsed_url._replace(
            path=re.sub(
                pattern=r'/v[1-2](/|$)?', repl=f'/v{self.API_VERSION}\\1', string=parsed_url.path, count=1
            )
        )
        self.parsed_url = util.url_path_join(parsed_url, self.API_PATH, trailing_slash=False)
        self.url = util.url_path_join(self.parsed_url, trailing_slash=True).geturl()

    @property
    def model(self) -> Type[TAirModel]:
        """Returns the respective model for this API."""

        return get_args(self.__orig_bases__[0])[0]

    def get(self, id: str, **kwargs) -> TAirModel:
        """
        Get an existing instance of a resource by ID.

        Arguments:
            id (str): Instance ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Raises:
        `AirUnexpectedResponse` - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organizations.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Organization NVIDIA 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """

        url = util.url_path_join(self.parsed_url, id, trailing_slash=True).geturl()
        response = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(response)

        return self.model(self, **response.json())

    def list(self, **kwargs) -> List[TAirModel]:
        """
        List existing instances of a resource.

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Raises:
        `AirUnexpectedResponse` - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organizations.list()
        [<Organization NVIDIA c51b49b6-94a7-4c93-950c-e7fa4883591>, <Organization Customer 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """

        url = util.url_path_join(self.parsed_url, trailing_slash=True).geturl()
        response = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(response, data_type=(list, dict))

        # response can either be a list of instances or a paginated response containing the first page of instances
        # TODO iterate pages
        parsed_response: Union[List[Dict], Dict[Literal['results'], List[Dict]]] = response.json()
        if isinstance(parsed_response, list):
            models_data = parsed_response
        else:
            models_data = parsed_response['results']

        return [self.model(self, **model_data) for model_data in models_data]

    def create(self, **kwargs) -> TAirModel:
        """
        Create a new instance of a resource.

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Raises:
        `AirUnexpectedResponse` - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.organizations.create(name='NVIDIA', members=[{'username': 'user1@nvidia.com', 'roles': ['Organization Admin']}, {'username': 'user2@nvidia.com'}])
        <Organization NVIDIA 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        """

        url = util.url_path_join(self.parsed_url, trailing_slash=True).geturl()
        response = self.client.post(url, json=kwargs)
        util.raise_if_invalid_response(response, status_code=HTTPStatus.CREATED)

        return self.model(self, **response.json())



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Topologyfile module
"""

from . import util
from .air_model import AirModel


class TopologyFile(AirModel):
    """A text file that describes a network topology."""

    _deletable = False
    _updatable = False

    def __repr__(self):
        return f'<TopologyFile {self.id}>'


class TopologyFileApi:
    """High-level interface for the TopologyFile API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url.replace('v1', 'v2') + '/topology-file/'

    def get(self, file_id: str, **kwargs) -> TopologyFile:
        """
        Get an existing topology file

        Arguments:
            file_id (str): TopologyFile ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        TopologyFile

        Raises:
        AirUnexpectedResponse - API did not return a 200 OK or valid response JSON

        Example:
        ```
        >>> air.topology_files.get('5cec8f3b-f449-47a3-a6ee-c5b81bf92ccf')
        <TopologyFile 5cec8f3b-f449-47a3-a6ee-c5b81bf92ccf>
        ```
        """
        url = f'{self.url}{file_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return TopologyFile(self, **res.json())

    # TODO: v2 list APIs require pagination handling
    # def list(self, **kwargs) -> 'list[TopologyFile]':
    #     """
    #     List existing TopologyFiles

    #     Arguments:
    #         kwargs (dict, optional): All other optional keyword arguments are applied as query
    #             parameters/filters

    #     Returns:
    #     list[TopologyFile]

    #     Raises:
    #     AirUnexpectedResponse - API did not return a 200 OK or valid response JSON

    #     Example:
    #     ```
    #     >>> air.topology_files.list()
    #     [<TopologyFile 5cec8f3b-f449-47a3-a6ee-c5b81bf92ccf>, <TopologyFile 811490b5-d2ef-4775-948a-2c1fbe0ce939>]
    #     ```
    #     """



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from typing import Optional, Union

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.accounts import Account
from air_sdk.v2.endpoints.organizations import Organization
from air_sdk.v2.air_model import AirModel, BaseEndpointApi, DataDict, PrimaryKey
from air_sdk.v2.endpoints.resource_budgets import ResourceBudget
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class UserConfig(AirModel):
    id: str = field(repr=False)
    name: str
    kind: str
    owner: Optional[Account] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    owner_budget: Optional[ResourceBudget] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    organization: Optional[Organization] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    organization_budget: Optional[ResourceBudget] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    content: Optional[str] = field(metadata=AirModel.FIELD_LAZY, repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return UserConfigEndpointApi

    @validate_payload_types
    def update(self, name: Optional[str] = None, content: Optional[str] = None) -> None:
        """Update specific fields of the user config.."""
        data = {'name': name, 'content': content}
        data = {key: value for (key, value) in data.items() if value is not None}
        super().update(**data)

    @validate_payload_types
    def full_update(self, name: str, content: str) -> None:
        """Update all  fields of the cloud-init assignment."""
        super().full_update(
            name=name,
            content=content,
            owner=self.owner,
            owner_budget=self.owner_budget,
            kind=self.kind,
            organization=self.organization,
            organization_budget=self.organization_budget,
        )


class UserConfigEndpointApi(
    mixins.ListApiMixin[UserConfig],
    mixins.CreateApiMixin[UserConfig],
    mixins.GetApiMixin[UserConfig],
    mixins.PatchApiMixin[UserConfig],
    mixins.PutApiMixin[UserConfig],
    mixins.DeleteApiMixin,
    BaseEndpointApi[UserConfig],
):
    API_PATH = 'userconfigs'
    model = UserConfig

    @validate_payload_types
    def create(
        self,
        name: str,
        kind: str,
        content: str,
        organization: Optional[Union[Organization, PrimaryKey]] = None,
    ) -> UserConfig:
        payload: DataDict = {
            'name': name,
            'kind': kind,
            'content': content,
            'organization': organization,
        }
        return super().create(**payload)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Simulation module
"""

import io
import os

from . import user_preference, util
from .air_model import AirModel


class Simulation(AirModel):
    """
    Manage a Simulation

    ### json
    Returns a JSON string representation of the simulation

    ### refresh
    Syncs the simulation with all values returned by the API

    ### update
    Update the simulation with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    _deletable = False

    def __repr__(self):
        if self._deleted or not self.title:
            return super().__repr__()
        return f"<Simulation '{self.title}' {self.id}>"

    def create_service(self, name, interface, dest_port, **kwargs):
        """
        Create a new service for this simulation

        Arguments:
            name (str): Name of the service
            interface (str | `SimulationInterface`): Interface that the service should be created
                for. This can be provided in one of the following formats:
                - [`SimulationInterface`](/docs/simulation-interface) object
                - ID of a [`SimulationInterface`](/docs/simulation-interface)
                - String in the format of 'node_name:interface_name'
            dest_port (int): Service port number
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Service`](/docs/service)

        Example:
        ```
        >>> simulation.create_service('myservice', 'oob-mgmt-server:eth0', 22, service_type='ssh')
        <Service myservice cc18d746-4cf0-4dd3-80c0-e7df68bbb782>
        >>> simulation.create_service('myservice', simulation_interface, 22, service_type='ssh')
        <Service myservice 9603d0d5-5526-4a0f-91b8-a600010d0091>
        ```
        """
        service = self._api.client.services.create(
            simulation=self.id, name=name, interface=interface, dest_port=dest_port, **kwargs
        )
        self.refresh()
        return service

    def add_permission(self, email, **kwargs):
        """
        Adds permission for a given user to this simulation.

        Arguments:
            email (str): Email address of the user being given permission
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Permission`](/docs/permission)

        Example:
        ```
        >>> simulation.add_permission('mrobertson@nvidia.com', write_ok=True)
        <Permission 217bea68-7048-4262-9bbc-b98ab16c603e>
        ```
        """
        return self._api.client.permissions.create(email=email, simulation=self.id, **kwargs)

    @util.required_kwargs(['action'])
    def control(self, **kwargs):
        """
        Sends a control command to the simulation.

        Arguments:
            action (str): Control command
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        dict: Response JSON

        Example:
        ```
        >>> simulation.control(action='destroy')
        {'result': 'success'}
        ```
        """
        url = f'{self._api.url}{self.id}/control/'
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res)
        return res.json()

    def load(self):
        """Alias for `start()`"""
        self.start()

    def start(self):
        """Start/load the simulation"""
        self.control(action='load')
        self.refresh()

    def stop(self):
        """Alias for `store()`"""
        self.store()

    def store(self):
        """Store and power off the simulation"""
        self.control(action='store')
        self.refresh()

    def delete(self, **kwargs):
        """Delete the simulation"""
        self.control(action='destroy', **kwargs)
        self._deleted = True

    def preferences(self, **kwargs):
        """
        Returns your account preferences for the simulation

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`UserPreference`](/docs/userpreference)

        Example:
        ```
        >>> simulation.preferences()
        {"show": true}
        ```
        """
        res = self._api.client.get(
            f'{self._api.url.replace("v1", "v2")}{self.id}/preferences/', params=kwargs
        )
        util.raise_if_invalid_response(res)
        return user_preference.UserPreference(self._api, _model=self, _version_override='2', **res.json())


class SimulationApi:
    """High-level interface for the Simulation API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/simulation/'

    def _create_v1(self, **kwargs):
        return self.client.post(self.url, json=kwargs)

    def _create_v2(self, **kwargs):
        topology_data = kwargs['topology_data']
        if isinstance(kwargs['topology_data'], io.IOBase):
            topology_data = kwargs['topology_data'].read()
        elif os.path.isfile(kwargs['topology_data']):
            with open(kwargs['topology_data'], 'r', encoding='utf-8') as data_file:
                topology_data = data_file.read()
        return self.client.post(self.url.replace('v1', 'v2'), json={**kwargs, 'topology_data': topology_data})

    @util.deprecated('SimulationApi.list()')
    def get_simulations(self):  # pylint: disable=missing-function-docstring
        return self.list()

    @util.deprecated('SimulationApi.get()')
    def get_simulation(self, simulation_id):  # pylint: disable=missing-function-docstring
        return self.get(simulation_id)

    @util.deprecated('SimulationApi.create()')
    def create_simulation(self, **kwargs):  # pylint: disable=missing-function-docstring
        return self.create(**kwargs)

    @util.deprecated('Simulation.update()')
    def update_simulation(self, simulation_id, data):  # pylint: disable=missing-function-docstring
        sim = self.get(simulation_id)
        sim.update(**data)

    def duplicate(self, simulation, **kwargs):
        """
        Duplicate/clone an existing simulation

        Arguments:
            simulation (str | `Simulation`): Simulation or ID of the snapshot to be duplicated
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        ([`Simulation`](/docs/simulation), dict): Newly created simulation and response JSON

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.duplicate(simulation=simulation)
        <Simulation my_sim 5ff3f0dc-7db8-4938-8257-765c8e48623a>
        ```
        """
        sim = simulation
        if isinstance(sim, str):
            sim = self.get(simulation)
        kwargs['action'] = 'duplicate'
        response = sim.control(**kwargs)
        return Simulation(self, **response['simulation']), response

    @util.deprecated('Simulation.control()')
    def control(self, simulation_id, action, **kwargs):  # pylint: disable=missing-function-docstring
        sim = self.get(simulation_id)
        return sim.control(action=action, **kwargs)

    def get_citc_simulation(self):
        """
        Get the active CITC reference simulation

        Returns:
        [`Simulation`](/docs/simulation)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.get_citc_simulation()
        <Simulation my_sim b9125419-7c6e-41db-bba9-7d647d63943e>
        ```
        """
        url = self.url + 'citc/'
        res = self.client.get(url)
        util.raise_if_invalid_response(res)
        return Simulation(self, **res.json())

    def get(self, simulation_id, **kwargs):
        """
        Get an existing simulation

        Arguments:
            simulation_id (str): Simulation ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Simulation`](/docs/simulation)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Simulation my_sim 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{simulation_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Simulation(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing simulations

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.list()
        [<Simulation sim1 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Simulation sim2 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Simulation(self, **simulation) for simulation in res.json()]

    @util.required_kwargs([('topology', 'topology_data')])
    def create(self, **kwargs):
        """
        Create a new simulation. The caller must provide either `topology` or `topology_data`.

        Arguments:
            topology (str | `Topology`, optional): `Topology` or ID
            topology_data (str | fd, optional): Topology in DOT format. This can be passed as a string
                containing the raw DOT data, a path to the DOT file on your local disk,
                or as a file descriptor for a local file
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Simulation`](/docs/simulation)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulations.create(topology=topology, title='my_sim')
        <Simulation my_sim 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        >>> air.simulations.create(topology_data='/tmp/my_net.dot', organization=my_org)
        <Simulation my_sim c0a4c018-0b85-4439-979d-9814166aaeac>
        >>> air.simulations.create(topology_data='graph "my_sim" { "server1" [ function="server" os="generic/ubuntu2204"] }', organization=my_org)
        <Simulation my_sim b9c0c68e-d4bd-4e9e-8a49-9faf41efaf70>
        >>> air.simulations.create(topology_data=open('/tmp/my_net.dot', 'r', encoding='utf-8')), organization=my_org)
        <Simulation my_sim 86162934-baa7-4d9a-a826-5863f92b03ef>
        ```
        """
        util.validate_timestamps(
            'Simulation created', expires_at=kwargs.get('expires_at'), sleep_at=kwargs.get('sleep_at')
        )
        if kwargs.get('topology'):
            res = self._create_v1(**kwargs)
        else:
            res = self._create_v2(**kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        return Simulation(self, **res.json())



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT
import inspect
from dataclasses import Field, fields, is_dataclass
from datetime import datetime, timezone
from functools import wraps
from typing import Optional, get_type_hints, TypeVar, Callable, Any, cast
from urllib.parse import ParseResult, urlparse
from uuid import UUID

from air_sdk.v2.typing import type_check

F = TypeVar('F', bound=Callable[..., Any])


def join_urls(*args: str) -> str:
    return '/'.join(frag.strip('/') for frag in args) + '/'


def iso_string_to_datetime(iso: str) -> Optional[datetime]:
    try:
        return datetime.fromisoformat(iso.replace('Z', '+00:00'))
    except ValueError:
        return None


def datetime_to_iso_string(date: datetime) -> str:
    return date.astimezone(tz=timezone.utc).isoformat().replace('+00:00', 'Z')


def to_uuid(uuid: str) -> Optional[UUID]:
    try:
        return UUID(uuid, version=4)
    except ValueError:
        return None


def to_url(url: str) -> Optional[ParseResult]:
    try:
        parsed_url = urlparse(url)
        return parsed_url if all((parsed_url.scheme, parsed_url.netloc, parsed_url.path)) else None
    except AttributeError:
        return None


def is_dunder(name: str) -> bool:
    delimiter = '__'
    return name.startswith(delimiter) and name.endswith(delimiter)


def as_field(class_or_instance: object, name: str) -> Optional[Field]:  # type: ignore[type-arg]
    if is_dataclass(class_or_instance):
        try:
            return next(field for field in fields(class_or_instance) if field.name == name)
        except StopIteration:
            pass
    return None


def validate_payload_types(func: F) -> F:
    """A wrapper for validating the type of payload during create."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        hints = get_type_hints(func)

        sig = inspect.signature(func)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        for name, value in bound_args.arguments.items():
            if name in hints:
                expected_type = hints[name]
                if not type_check(value, expected_type):
                    raise TypeError(f"Argument '{name}' must be {expected_type}, got {type(value)}")

        return func(*args, **kwargs)

    return cast(F, wrapper)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Union

from air_sdk.v2.endpoints import mixins, MarketplaceDemoTag, Account
from air_sdk.v2.endpoints.simulations import Simulation
from air_sdk.v2.air_model import AirModel, BaseEndpointApi, PrimaryKey, DataDict
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class MarketplaceDemo(AirModel):
    id: str
    name: str
    owner_email: str = field(repr=False)
    created: datetime = field(repr=False)
    modified: datetime = field(repr=False)
    icon: Optional[str] = field(repr=False)
    liked_by_account: bool = field(repr=False)
    published: bool = field(repr=False)
    like_count: int = field(repr=False)
    snapshot: Simulation = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    repo: Optional[str] = field(repr=False)
    description: Optional[str] = field(repr=False)
    documentation: Optional[str] = field(repr=False)
    documentation_details: Optional[str] = field(metadata=AirModel.FIELD_LAZY, repr=False)
    tags: List[str] = field(default_factory=list, repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return MarketplaceDemoEndpointApi

    @validate_payload_types
    def update(
        self,
        name: Optional[str] = None,
        snapshot: Optional[Union[Simulation, PrimaryKey]] = None,
        description: Optional[str] = None,
        documentation: Optional[str] = None,
        repo: Optional[str] = None,
        icon: Optional[str] = None,
        owner: Optional[Union[Account, PrimaryKey]] = None,
        tags: Optional[List[Union[MarketplaceDemoTag, str]]] = None,
        published: Optional[bool] = None,
    ) -> None:
        data = {
            'name': name,
            'snapshot': snapshot,
            'description': description,
            'documentation': documentation,
            'repo': repo,
            'icon': icon,
            'owner': owner,
            'tags': tags,
            'published': published,
        }
        data = {key: value for (key, value) in data.items() if value is not None}
        super().update(**data)

    @validate_payload_types
    def full_update(
        self,
        name: str,
        snapshot: Union[Simulation, PrimaryKey],
        description: Optional[str],
        documentation: Optional[str],
        repo: Optional[str],
        icon: Optional[str],
        owner: Optional[Union[Account, PrimaryKey]],
        tags: List[Union[MarketplaceDemoTag, str]],
        published: bool = False,
    ) -> None:
        data = {
            'name': name,
            'snapshot': snapshot,
            'description': description,
            'documentation': documentation,
            'repo': repo,
            'icon': icon,
            'owner': owner,
            'tags': tags,
            'published': published,
        }
        super().full_update(**data)


class MarketplaceDemoEndpointApi(
    mixins.ListApiMixin[MarketplaceDemo],
    mixins.CreateApiMixin[MarketplaceDemo],
    mixins.GetApiMixin[MarketplaceDemo],
    mixins.PutApiMixin[MarketplaceDemo],
    mixins.PatchApiMixin[MarketplaceDemo],
    mixins.DeleteApiMixin,
    BaseEndpointApi[MarketplaceDemo],
):
    API_PATH = 'marketplace-demos'

    model = MarketplaceDemo

    @validate_payload_types
    def create(
        self,
        name: str,
        snapshot: Union[Simulation, PrimaryKey],
        description: Optional[str] = None,
        documentation: Optional[str] = None,
        repo: Optional[str] = None,
        icon: Optional[str] = None,
        owner: Optional[Union[Account, PrimaryKey]] = None,
        tags: Optional[List[Union[MarketplaceDemoTag, str]]] = None,
        published: bool = False,
    ) -> MarketplaceDemo:
        payload: DataDict = {
            'snapshot': snapshot,
            'documentation': documentation,
            'repo': repo,
            'icon': icon,
            'name': name,
            'description': description,
            'published': published,
        }
        for key, value in (
            ('owner', owner),
            ('tags', tags),
        ):
            if value:
                payload[key] = value
        return super().create(**payload)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

# ruff:noqa: F401

__all__ = [
    'AccountEndpointApi',
    'Account',
    'AnnouncementEndpointApi',
    'Announcement',
    'ApiTokenEndpointApi',
    'ApiToken',
    'CloudInitEndpointApi',
    'CloudInit',
    'ImageEndpointApi',
    'Image',
    'InterfaceEndpointApi',
    'Interface',
    'JobEndpointApi',
    'Job',
    'Link',
    'LinkEndpointApi',
    'MarketplaceDemoEndpointApi',
    'MarketplaceDemo',
    'MarketplaceDemoTagsEndpointApi',
    'MarketplaceDemoTag',
    'NodeEndpointApi',
    'Node',
    'OrganizationEndpointApi',
    'Organization',
    'ResourceBudgetEndpointApi',
    'ResourceBudget',
    'ServiceEndpointApi',
    'Service',
    'SimulationEndpointApi',
    'Simulation',
    'SystemEndpointApi',
    'System',
    'UserConfigEndpointApi',
    'UserConfig',
    'WorkerEndpointApi',
    'Worker',
    'mixins',
]

from air_sdk.v2.endpoints.accounts import Account, AccountEndpointApi
from air_sdk.v2.endpoints.announcements import Announcement, AnnouncementEndpointApi
from air_sdk.v2.endpoints.api_tokens import ApiToken, ApiTokenEndpointApi
from air_sdk.v2.endpoints.images import Image, ImageEndpointApi
from air_sdk.v2.endpoints.interfaces import Interface, InterfaceEndpointApi, Link, LinkEndpointApi
from air_sdk.v2.endpoints.jobs import Job, JobEndpointApi
from air_sdk.v2.endpoints.marketplace_demo_tags import MarketplaceDemoTag, MarketplaceDemoTagsEndpointApi
from air_sdk.v2.endpoints.marketplace_demos import MarketplaceDemo, MarketplaceDemoEndpointApi
from air_sdk.v2.endpoints.nodes import Node, NodeEndpointApi
from air_sdk.v2.endpoints.cloud_inits import CloudInit, CloudInitEndpointApi
from air_sdk.v2.endpoints.organizations import Organization, OrganizationEndpointApi
from air_sdk.v2.endpoints.resource_budgets import ResourceBudget, ResourceBudgetEndpointApi
from air_sdk.v2.endpoints.services import Service, ServiceEndpointApi
from air_sdk.v2.endpoints.simulations import Simulation, SimulationEndpointApi
from air_sdk.v2.endpoints.systems import System, SystemEndpointApi
from air_sdk.v2.endpoints.user_configs import UserConfig, UserConfigEndpointApi
from air_sdk.v2.endpoints.workers import Worker, WorkerEndpointApi
from air_sdk.v2.endpoints import mixins



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT
from __future__ import annotations

import json
from http import HTTPStatus
from typing import Optional, Any, TypedDict, List, Callable, Iterator, TYPE_CHECKING, Dict, Generic

from air_sdk.v2.air_json_encoder import AirJSONEncoder
from air_sdk.v2.air_model import PrimaryKey, DataDict, TAirModel_co
from air_sdk.v2.utils import join_urls
from air_sdk.util import raise_if_invalid_response

if TYPE_CHECKING:
    from air_sdk.v2 import AirApi


def serialize_payload(data: Dict[str, Any] | List[Dict[str, Any]]) -> str:
    """Serialize the dictionary of values into json using the AirJSONEncoder."""
    return json.dumps(data, indent=None, separators=(',', ':'), cls=AirJSONEncoder)


class BaseApiMixin:
    """A base class for API Mixins.

    This is primarily used for type hinting.
    """

    __api__: AirApi
    url: str
    load_model: Callable[[DataDict], TAirModel_co]


class PaginatedResponseData(TypedDict):
    count: int
    next: Optional[str]
    previous: Optional[str]
    results: List[DataDict]


class ListApiMixin(BaseApiMixin, Generic[TAirModel_co]):
    """Returns an iterable of model objects.

    Handles pagination in the background.
    """

    def list(self, **params: Any) -> Iterator[TAirModel_co]:
        """Return an iterator of model instances."""
        url = self.url
        # Set up pagination
        next_url = None
        params.setdefault('limit', self.__api__.client.pagination_page_size)
        params = json.loads(serialize_payload(params))  # Accounts for UUIDs and AirModel params
        while url or next_url:
            if isinstance(next_url, str):
                response = self.__api__.client.get(next_url)
            else:
                response = self.__api__.client.get(url, params=params)
            raise_if_invalid_response(response)
            paginated_response_data: PaginatedResponseData = response.json()
            url = None  # type: ignore[assignment]
            next_url = paginated_response_data['next']
            for obj_data in paginated_response_data['results']:
                yield self.load_model(obj_data)


class CreateApiMixin(BaseApiMixin, Generic[TAirModel_co]):
    def create(self, *args: Any, **kwargs: Any) -> TAirModel_co:
        response = self.__api__.client.post(self.url, data=serialize_payload(kwargs))
        raise_if_invalid_response(response, status_code=HTTPStatus.CREATED)
        return self.load_model(response.json())


class GetApiMixin(BaseApiMixin, Generic[TAirModel_co]):
    def get(self, pk: PrimaryKey, **params: Any) -> TAirModel_co:
        detail_url = join_urls(self.url, str(pk))
        response = self.__api__.client.get(detail_url, params=params)
        raise_if_invalid_response(response)
        return self.load_model(response.json())


class PutApiMixin(BaseApiMixin, Generic[TAirModel_co]):
    def put(self, pk: PrimaryKey, **kwargs: Any) -> TAirModel_co:
        response = self.__api__.client.put(join_urls(self.url, str(pk)), data=serialize_payload(kwargs))
        raise_if_invalid_response(response, status_code=HTTPStatus.OK)
        return self.load_model(response.json())


class PatchApiMixin(BaseApiMixin, Generic[TAirModel_co]):
    def patch(self, pk: PrimaryKey, **kwargs: Any) -> TAirModel_co:
        response = self.__api__.client.patch(join_urls(self.url, str(pk)), data=serialize_payload(kwargs))
        raise_if_invalid_response(response, status_code=HTTPStatus.OK)
        return self.load_model(response.json())


class DeleteApiMixin(BaseApiMixin):
    def delete(self, pk: PrimaryKey, **kwargs: Any) -> None:
        """Deletes the instances with the specified primary key."""
        detail_url = join_urls(self.url, str(pk))
        response = self.__api__.client.delete(detail_url, json=kwargs)
        raise_if_invalid_response(response, status_code=HTTPStatus.NO_CONTENT, data_type=None)



# SPDX-FileCopyrightText: Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
UserPreference module
"""

import json
import re

from .air_model import AirModel
from . import util


class UserPreference(AirModel):
    """
    A collection of your user preferences, which may be global for your account or specific to a single
    simulation.

    ### json
    Returns a JSON string representation of the preferences object

    ### refresh
    Syncs the key with all values returned by the API
    """

    _deletable = False

    def __init__(self, api, **kwargs):
        _model = kwargs.pop('_model', None)
        _version_override = kwargs.pop('_version_override', None)
        super().__init__(api, **kwargs)
        self._model = _model
        self._version_override = _version_override
        self._url = self._build_url()

    def __repr__(self):
        return json.dumps(self.preferences)

    def __setattr__(self, name, value):
        if not getattr(self, '_url', None):
            return super().__setattr__(name, value)
        if self.preferences.get(name) != value:
            res = self._api.client.patch(self._url, json={name: value})
            util.raise_if_invalid_response(res)
        self.preferences[name] = value
        return None

    def _build_url(self):
        url = self._api.url
        if self._model:
            url += f'{self._model.id}/'
        url += 'preferences/'
        if self._version_override:
            url = re.sub(r'/v\d/', f'/v{self._version_override}/', url)
        return url

    def refresh(self):
        """Syncs the object with all values returned by the API"""
        instance = self._model or self._api
        self._load(**instance.preferences().__dict__)

    def update(self, **kwargs):
        """
        Update the object with the provided data

        Arguments:
            kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
        """
        self.refresh()
        self.preferences.update(kwargs)
        res = self._api.client.put(self._url, json=self.__dict__)
        util.raise_if_invalid_response(res)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import json
from dataclasses import dataclass, field
from http import HTTPStatus
from typing import Optional, List, Dict, Union, Any, Iterator, Type, TYPE_CHECKING

from air_sdk.util import raise_if_invalid_response

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.mixins import serialize_payload

from air_sdk.v2.endpoints.nodes import Node
from air_sdk.v2.endpoints.simulations import Simulation
from air_sdk.v2.air_model import AirModel, BaseEndpointApi, PrimaryKey, DataDict, BaseModel

from air_sdk.v2.utils import validate_payload_types, join_urls

if TYPE_CHECKING:
    from air_sdk.v2 import AirApi


@dataclass(eq=False)
class Interface(AirModel):
    id: str
    name: str
    link_up: bool
    internal_ipv4: Optional[str]
    full_ipv6: Optional[str]
    prefix_ipv6: Optional[str]
    interface_type: Optional[str]
    port_number: int
    preserve_mac: bool
    outbound: bool
    mac_address: Optional[str]
    node: Node = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    simulation: Simulation = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    link: Optional[Link] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)

    def __post_init__(self, _api: AirApi) -> None:
        super().__post_init__(_api)
        if self.link is not None:
            self.link.__simulation_pk__ = self.simulation.__pk__  # type: ignore[attr-defined]

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return InterfaceEndpointApi

    @validate_payload_types
    def update(
        self,
        name: Optional[str] = None,
        port_number: Optional[int] = None,
        outbound: Optional[bool] = None,
        preserve_mac: Optional[bool] = None,
        link_up: Optional[bool] = None,
        link: Optional[Union[Link, PrimaryKey]] = None,
    ) -> None:
        data = {
            'name': name,
            'link_up': link_up,
            'port_number': port_number,
            'outbound': outbound,
            'preserve_mac': preserve_mac,
            'link': link,
        }
        data = {key: value for (key, value) in data.items() if value is not None}
        super().update(**data)

    @validate_payload_types
    def full_update(
        self,
        name: str,
        port_number: int,
        outbound: bool,
        preserve_mac: bool,
        link_up: bool,
        link: Optional[Union[Link, PrimaryKey]],
    ) -> None:
        data = {
            'name': name,
            'link_up': link_up,
            'port_number': port_number,
            'outbound': outbound,
            'preserve_mac': preserve_mac,
            'link': link,
        }
        super().update(**data)


class InterfaceEndpointApi(
    mixins.ListApiMixin[Interface],
    mixins.CreateApiMixin[Interface],
    mixins.GetApiMixin[Interface],
    mixins.PatchApiMixin[Interface],
    mixins.DeleteApiMixin,
    BaseEndpointApi[Interface],
):
    API_PATH = 'simulations/nodes/interfaces/'
    model = Interface

    @validate_payload_types
    def create(
        self,
        name: str,
        node: Union[Node, PrimaryKey],
        interface_type: str = 'ETH_INTF',
        link_up: bool = False,
        port_number: int = 0,
        outbound: bool = False,
        preserve_mac: bool = False,
        mac_address: Optional[str] = None,
        internal_ipv4: Optional[str] = None,
        full_ipv6: Optional[str] = None,
        prefix_ipv6: Optional[str] = None,
        link: Optional[Union[Link, PrimaryKey]] = None,
    ) -> Interface:
        return super().create(
            name=name,
            node=node,
            interface_type=interface_type,
            link_up=link_up,
            port_number=port_number,
            outbound=outbound,
            preserve_mac=preserve_mac,
            internal_ipv4=internal_ipv4,
            full_ipv6=full_ipv6,
            prefix_ipv6=prefix_ipv6,
            mac_address=mac_address,
            link=link,
        )


"""
Link and Interface share a module to prevent circular import issues.
"""

LinkDataList = List[Dict[str, List[Union[PrimaryKey, Interface]]]]


@dataclass(eq=False)
class Link(AirModel):
    id: str
    simulation_interfaces: List[Interface] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)

    def __refresh__(self, refreshed_obj: Optional[BaseModel] = None) -> None:
        link_api = self.get_model_api()(self.__api__)
        refreshed_obj = next(link_api.list(simulation=self.__simulation_pk__, ids=str(self.__pk__)))
        super().__refresh__(refreshed_obj)

    @classmethod
    def get_model_api(cls) -> Type[LinkEndpointApi]:
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return LinkEndpointApi


class LinkEndpointApi(mixins.ListApiMixin[Link], BaseEndpointApi[Link]):
    API_PATH = 'simulations/nodes/interfaces/links/'
    model = Link

    def list(self, simulation: Union[Simulation, PrimaryKey], **params: Any) -> Iterator[Link]:  # type: ignore[override]
        if simulation is None:
            raise TypeError('simulation may not be `None`.')
        return super().list(simulation=simulation, **params)

    @validate_payload_types
    def bulk_create(self, simulation: Union[Simulation, PrimaryKey], links: LinkDataList) -> List[Link]:
        payload = {'simulation': simulation, 'links': links}
        if len(links) == 0:
            raise ValueError('links must be a non-empty list of dictionaries.')
        for i, link in enumerate(links):
            # Validate the payload format and make content json serializable.
            sim_iface_key = 'simulation_interfaces'
            if set(link.keys()) != {sim_iface_key}:
                raise ValueError(
                    'Expected link to have format {"simulation_interfaces": [interface_1, interface_2]} '
                    f'but got {link}'
                )

        response = self.__api__.client.post(
            join_urls(self.url, 'bulk-create'), data=serialize_payload(payload)
        )
        raise_if_invalid_response(response, status_code=HTTPStatus.CREATED)
        response_data: DataDict = response.json()
        link_data_list = response_data['links']
        return [self.load_model(link_data) for link_data in link_data_list]

    @validate_payload_types
    def bulk_delete(
        self, simulation: Union[Simulation, PrimaryKey], links: Optional[List[Union[Link, PrimaryKey]]] = None
    ) -> Dict[str, int]:
        if not simulation:
            raise ValueError('`simulation` must be specified')
        if isinstance(links, list) and len(links) == 0:
            raise ValueError('If `links` is provided it must be a non-empty list.')
        params = json.loads(serialize_payload({'simulation': simulation, 'links': links}))
        link_id_list = params.pop('links')
        if isinstance(link_id_list, list):
            params['ids'] = ','.join(link_id_list)
        response = self.__api__.client.delete(join_urls(self.url, 'bulk-delete'), params=params)
        raise_if_invalid_response(response, status_code=HTTPStatus.OK)
        links_deleted: int = response.json()['links_deleted']
        return {'links_deleted': links_deleted}



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""Exposes the AIR API client module"""

from .air_api import *  # noqa: F403
from . import v2  # noqa: F401



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Union, List, Dict

from air_sdk.util import raise_if_invalid_response
from air_sdk.v2.endpoints import mixins
from air_sdk.v2.air_model import AirModel, BaseEndpointApi
from air_sdk.v2.endpoints.mixins import serialize_payload
from air_sdk.v2.utils import join_urls, validate_payload_types


@dataclass(eq=False)
class Worker(AirModel):
    id: str = field(repr=False)
    fqdn: str
    cpu_arch: str
    available: bool = field(repr=False)
    capabilities: str = field(repr=False)
    cpu: int = field(repr=False)
    contact: Union[str, List[str], Dict[str, str]] = field(repr=False)
    created: datetime = field(repr=False)
    fleet: str = field(repr=False)
    gpu: int = field(repr=False)
    ip_address: str = field(repr=False)
    memory: int = field(repr=False)
    modified: datetime = field(repr=False)
    port_range: str = field(repr=False)
    registered: bool = field(repr=False)
    storage: int = field(repr=False)
    tunnel_port: int = field(repr=False)
    vgpu: int = field(repr=False)

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return WorkerEndpointApi

    def heartbeat(self) -> None:
        """Keeps the worker alive with via a heartbeat to the manager."""
        heartbeat_url = join_urls(self.__api__.workers.url, 'heartbeat')
        response = self.__api__.client.get(heartbeat_url)
        raise_if_invalid_response(response, data_type=None)

    @validate_payload_types
    def update_inventory(
        self,
        airstrike_version: Optional[str] = None,
        architecture: Optional[str] = None,
        docker: Optional[str] = None,
        kernel: Optional[str] = None,
        libvirt: Optional[str] = None,
        operating_system: Optional[str] = None,
        proxy_image: Optional[str] = None,
        worker_version: Optional[str] = None,
    ) -> None:
        """Update the worker's inventory with the manager."""
        data = {
            'airstrike_version': airstrike_version,
            'architecture': architecture,
            'docker': docker,
            'kernel': kernel,
            'libvirt': libvirt,
            'operating_system': operating_system,
            'proxy_image': proxy_image,
            'worker_version': worker_version,
        }
        inventory_url = join_urls(self.__api__.workers.url, 'inventory')
        response = self.__api__.client.put(
            inventory_url,
            data=serialize_payload({key: value for (key, value) in data.items() if value is not None}),
        )
        raise_if_invalid_response(response, data_type=None)


class WorkerEndpointApi(
    mixins.ListApiMixin[Worker],
    mixins.GetApiMixin[Worker],
    BaseEndpointApi[Worker],
):
    API_PATH = 'workers'
    model = Worker



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
SimulationNode module
"""

import json
from functools import wraps
from typing import Callable, Dict, Optional, TypeVar, Literal, Union

from . import util
from .air_model import AirModel, AirModelAPI
from .userconfig import UserConfig

CloudInitAssignmentFields = Literal['user_data', 'meta_data']
CloudInitAssignmentResponseFields = Union[
    CloudInitAssignmentFields, Literal['simulation_node', 'user_data_name', 'meta_data_name']
]
CloudInitAssignment = Dict[CloudInitAssignmentFields, Optional[str]]
CloudInitAssignmentResponse = Dict[CloudInitAssignmentResponseFields, Optional[str]]


class SimulationNode(AirModel):
    """
    Manage a SimulationNode

    ### json
    Returns a JSON string representation of the simulation node

    ### refresh
    Syncs the simulation node with all values returned by the API

    ### update
    Update the simulation node with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    _deletable = False

    def __init__(self, api, **kwargs):
        super().__init__(api, **kwargs)

        # inject v2 functionalities for ease of use
        self_v2_api = _SimulationNodeAPI(self._api.client)
        self_v2 = self_v2_api.model(self_v2_api)

        TMethod = TypeVar('TMethod', bound='Callable')

        def _with_updated_v2(method: TMethod) -> TMethod:
            """Updates mirrored v2 SimulatedNode with v1 data."""

            @wraps(method)
            def _wrapper(*args, **kwargs):
                self_v2._load(**json.loads(self.json()))
                return method(*args, **kwargs)

            return _wrapper

        self.get_cloud_init_assignment = _with_updated_v2(self_v2.get_cloud_init_assignment)
        self.set_cloud_init_assignment = _with_updated_v2(self_v2.set_cloud_init_assignment)

    def __repr__(self):
        if self._deleted:
            return super().__repr__()
        return f'<SimulationNode {self.id}>'

    @util.required_kwargs(['executor', 'data'])
    def create_instructions(self, **kwargs):
        """
        Create instructions for the `SimulationNode`'s agent to execute

        Arguments:
            data (str | list): Instruction data
            executor (str): Agent executor type
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        dict: Response JSON

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> simulation_node.create_instructions(data='echo foo', executor='shell')
        {'id': '67f73552-ffdf-4e5f-9881-aeae227604a3'}
        ```
        """
        url = f'{self._api.url}{self.id}/instructions/'
        if isinstance(kwargs['data'], list):
            kwargs['data'] = '\n'.join(kwargs['data'])
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201, data_type=str)
        return {'id': res.json()}

    def list_instructions(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List all instructions for a `SimulationNode`

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> simulation_node.instructions.list()
        [{'id': '56abc69b-489f-429a-aed9-600f26afc956'}, {'id': '7c9c3449-f071-4bbc-bb42-bef04e44d74e'}]
        ```
        """  # pylint: enable=line-too-long
        url = f'{self._api.url}{self.id}/instructions/'
        res = self._api.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return res.json()

    def delete_instructions(self):
        """
        Delete all instructions for a `SimulationNode`

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Instruction delete failed

        Example:
        ```
        >>> simulation_node.instructions.delete()
        ```
        """
        url = f'{self._api.url}{self.id}/instructions/'
        res = self._api.client.delete(url)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)

    @util.required_kwargs(['action'])
    def control(self, **kwargs):
        """
        Sends a control command to the `SimulationNode`.

        Arguments:
            action (str): Control command
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        dict: Response JSON

        Example:
        ```
        >>> simulation_node.control(action='reset')
        {'result': 'success'}
        ```
        """
        url = f'{self._api.url}{self.id}/control/'
        res = self._api.client.post(url, json=kwargs)
        util.raise_if_invalid_response(res)
        return res.json()

    def rebuild(self, **kwargs):
        """
        Rebuild the `SimulationNode` back to it's initial state. **All existing data will be lost!**
        """
        self.control(action='rebuild', **kwargs)

    def reset(self, **kwargs):
        """Reset the `SimulationNode`"""
        self.control(action='reset', **kwargs)


class SimulationNodeApi:
    """Wrapper for the SimulationNode API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/simulation-node/'

    @util.deprecated('SimulationNode.update()')
    def update_simulation_node(self, simulation_node_id, data):  # pylint: disable=missing-function-docstring
        node = self.get(simulation_node_id)
        node.update(**data)

    @util.deprecated('SimulationNodeApi.list()')
    def get_simulation_nodes(self, **kwargs):  # pylint: disable=missing-function-docstring
        return self.list(**kwargs)

    def get(self, simulation_node_id, **kwargs):
        """
        Get an existing simulation node

        Arguments:
            simulation_node_id (str): SimulationNode ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`SimulationNode`](/docs/simulationnode)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulation_nodes.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <SimulationNode my_sim 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{simulation_node_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return SimulationNode(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing simulation nodes

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.simulation_nodes.list()
        [<SimulationNode sim1 c51b49b6-94a7-4c93-950c-e7fa4883591>, <SimulationNode sim2 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [SimulationNode(self, **simulation_node) for simulation_node in res.json()]


class _v2:
    """Temporary private class for v2 related SimulationNode resources."""

    class SimulationNode(AirModel):
        """Manage a v2 SimulationNode."""

        CLOUD_INIT_PATH = 'cloud-init'

        def get_cloud_init_assignment(self) -> CloudInitAssignmentResponse:
            """Returns current state of cloud-init script assignments for the node."""

            url = util.url_path_join(
                self._api.parsed_url, self.id, self.CLOUD_INIT_PATH, trailing_slash=True
            ).geturl()
            response = self._api.client.get(url)
            util.raise_if_invalid_response(response, data_type=dict)

            return response.json()

        def set_cloud_init_assignment(
            self, script_mapping: CloudInitAssignment
        ) -> CloudInitAssignmentResponse:
            """
            Edits cloud-init script assignment for the node as defined in `script_mapping`.
            Returns new state of cloud-init script assignments for the node.

            Any combination of script keys can be provided within `script_mapping`.
            Explicit `None` as a value for a key will clear the assignment for that specific script.

            Example:
            ```
            # only sets user-data
            node.set_cloud_init_assignment({
                'user_data': my_script
            })

            # sets user-data, clears assignment for meta-data
            node.set_cloud_init_assignment({
                'user_data': my_script,
                'meta_data': None
            })

            # identical to `node.get_cloud_init_assignment()`
            node.set_cloud_init_assignment({})
            ```
            """

            patch_payload: CloudInitAssignment = {}
            for script_type in [
                script_type for script_type in script_mapping if script_type in ['user_data', 'meta_data']
            ]:
                script = script_mapping[script_type]
                if isinstance(script, UserConfig):
                    patch_payload[script_type] = script.id
                else:
                    patch_payload[script_type] = script

            if not patch_payload:
                return self.get_cloud_init_assignment()

            url = util.url_path_join(
                self._api.parsed_url, self.id, self.CLOUD_INIT_PATH, trailing_slash=True
            ).geturl()
            response = self._api.client.patch(url, json=patch_payload)
            util.raise_if_invalid_response(response, data_type=dict)

            return response.json()


class _SimulationNodeAPI(AirModelAPI[_v2.SimulationNode]):
    """High-level interface for the SimulationNode v2 API."""

    API_VERSION = 2
    API_PATH = 'simulations/nodes'



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.air_model import AirModel, BaseEndpointApi


@dataclass(eq=False)
class MarketplaceDemoTag(AirModel):
    name: str

    @property
    def primary_key_field(self) -> str:
        return 'name'

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return MarketplaceDemoTagsEndpointApi


class MarketplaceDemoTagsEndpointApi(
    mixins.ListApiMixin[MarketplaceDemoTag], BaseEndpointApi[MarketplaceDemoTag]
):
    API_PATH = 'marketplace-demo-tags'

    model = MarketplaceDemoTag



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
UserConfig model related module.
"""

import json
import os
from io import TextIOBase
from pathlib import Path
from typing import Dict, Optional, TextIO, Union

from .air_model import AirModel, AirModelAPI
from .exceptions import AirObjectDeleted
from .organization import Organization


class UserConfig(AirModel):
    """
    Manage a UserConfig.

    ### delete
    Delete the UserConfig. Once successful, the object should no longer be used and will raise
    `AirDeletedObject` when referenced.

    Raises:
    `AirUnexpectedResponse` - Delete failed

    ### json
    Returns a JSON string representation of the UserConfig.

    ### refresh
    Syncs the UserConfig with all values returned by the API

    ### update
    Update the UserConfig with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    KIND_CLOUD_INIT_USER_DATA = 'cloud-init-user-data'
    KIND_CLOUD_INIT_META_DATA = 'cloud-init-meta-data'
    VALID_KINDS = (KIND_CLOUD_INIT_USER_DATA, KIND_CLOUD_INIT_META_DATA)

    def __repr__(self):
        try:
            self_dict: Dict = json.loads(self.json())
            if self_dict.get('name', False) and self_dict.get('kind', False):
                return f'<UserConfig {self.name} {self.kind} {self.id}>'
        except AirObjectDeleted:
            pass

        return super().__repr__()


class UserConfigAPI(AirModelAPI[UserConfig]):
    """High-level interface for the UserConfig API."""

    API_VERSION = 2
    API_PATH = 'userconfigs'

    def create(
        self,
        name: str,
        kind: str,
        organization: Optional[Union[Organization, str]],
        content: Union[str, Path, TextIO],
    ) -> UserConfig:
        """
        Create a new UserConfig. Content data can be provided as a plain string, path to an existing file or an open file handle.
        Keep in mind that:
        - When passing a file path, it will be opened for reading using default encoding
        - When passing a file handle, it is assumed to be opened using a proper encoding and will be read from as-is
        - Due to a small size of UserConfig scripts, content will be loaded into memory in its entirety

        Arguments:
            name: UserConfig name
            kind: UserConfig kind, must be one of `UserConfig.VALID_KINDS`
            organization: Organization instance / ID to create the UserConfig in
            content: UserConfig data
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Raises:
        `AirUnexpectedResponse` - API did not return a 200 OK
            or valid response JSON
        `AttributeError` - provided content object is not one of the allowed types
        `FileNotFoundError` - when providing a path to a content file and the file is not present

        Example:
        ```
        >>> air.user_configs.create(name='my-config', kind=air.user_configs.model.KIND_CLOUD_INIT_USER_DATA, organization=my_org, content="my-content")
        <UserConfig my-config cloud-init-user-data 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        """

        if isinstance(content, str):
            if os.path.exists(content):
                with open(content, 'r') as content_file:
                    parsed_content = content_file.read()
            else:
                parsed_content = content
        elif isinstance(content, Path):
            with content.open('r') as content_file:
                parsed_content = content_file.read()
        elif isinstance(content, TextIOBase):
            parsed_content = content.read()
        else:
            raise AttributeError(f'Unexpected content type provided: `{type(content)}`')

        return super().create(
            name=name,
            kind=kind,
            organization=organization,
            content=parsed_content,
        )



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from typing import Dict, Any, Optional
from json import JSONDecodeError
from urllib.parse import urlparse

import requests  # type: ignore[import-untyped]

from air_sdk.v2.utils import join_urls
from air_sdk import logger, const
from air_sdk.air_api import AirAuthorizationError, _normalize_api_url  # type: ignore[attr-defined]


class Client(requests.Session):  # type: ignore
    """A session client for managing the execution of API requests."""

    def __init__(self, api_url: str):
        super().__init__()
        self.headers.update({'content-type': 'application/json', 'Authorization': None})
        self.api_url = _normalize_api_url(api_url)
        self.base_url = join_urls(self.api_url, 'v2')
        self.connect_timeout = const.DEFAULT_CONNECT_TIMEOUT
        self.read_timeout = const.DEFAULT_READ_TIMEOUT
        self.pagination_page_size = const.DEFAULT_PAGINATION_PAGE_SIZE

    def rebuild_auth(self, prepared_request, response):
        """Allow credential sharing between nvidia.com and cumulusnetworks.com only"""
        if urlparse(prepared_request.url).hostname in const.ALLOWED_HOSTS:
            return
        super().rebuild_auth(prepared_request, response)

    def request(self, method: str, url: str, **kwargs: Dict[str, Any]) -> requests.Response:
        """Override request method to pass the timeout"""
        kwargs.setdefault('timeout', (self.connect_timeout, self.read_timeout))  # type: ignore[arg-type]
        return super().request(method, url, **kwargs)

    def authenticate(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        bearer_token: Optional[str] = None,
    ) -> None:
        if bearer_token is None and not (username and password):
            raise ValueError(
                'Unable to authenticate client. Please provide one of the following: '
                '(1) `username` and `password`; '
                '(2) `bearer_token`'
            )
        token = bearer_token or self.get_token(username, password)  # type: ignore[arg-type]
        self.headers.update({'authorization': f'Bearer {token}'})

    def get_token(self, username: str, password: str) -> str:
        login_url = join_urls(self.api_url, 'v1', 'login')
        response = self.post(login_url, json={'username': username, 'password': password})
        try:
            token = response.json().get('token', None)
            if isinstance(token, str):
                return token
            logger.debug('AirApi.get_token :: Response JSON')  # type: ignore
            logger.debug(response.json())  # type: ignore
            raise AirAuthorizationError('API did not provide a token for ' + username)
        except JSONDecodeError:
            raise AirAuthorizationError('API did not return a valid JSON response')
